"""image utils for marigold"""
import math
import torch
from torchvision.transforms import InterpolationMode
from torchvision.transforms.functional import resize

def resize_max_res(img: torch.Tensor, max_edge_resolution: int, resample_method: InterpolationMode) -> torch.Tensor:
    """Resize image to limit maximum edge length while keeping aspect ratio."""
    assert 4 == img.dim(), f"Invalid input shape {img.shape}"

    original_height, original_width = img.shape[-2:]
    downscale_factor = min(max_edge_resolution / original_width, max_edge_resolution / original_height)

    new_width = int(original_width * downscale_factor)
    new_height = int(original_height * downscale_factor)

    resized_img = resize(img, (new_height, new_width), resample_method, antialias=True)
    return resized_img


def find_batch_size(ensemble_size: int, input_res: int, dtype: torch.dtype) -> int:
    """Automatically search for suitable operating batch size."""

    # Search table for suggested max. inference batch size
    bs_search_table = [
        # tested on A100-PCIE-80GB
        {"res": 768, "total_vram": 79, "bs": 35, "dtype": torch.float32},
        {"res": 1024, "total_vram": 79, "bs": 20, "dtype": torch.float32},
        # tested on A100-PCIE-40GB
        {"res": 768, "total_vram": 39, "bs": 15, "dtype": torch.float32},
        {"res": 1024, "total_vram": 39, "bs": 8, "dtype": torch.float32},
        {"res": 768, "total_vram": 39, "bs": 30, "dtype": torch.float16},
        {"res": 1024, "total_vram": 39, "bs": 15, "dtype": torch.float16},
        # tested on RTX3090, RTX4090
        {"res": 512, "total_vram": 23, "bs": 20, "dtype": torch.float32},
        {"res": 768, "total_vram": 23, "bs": 7, "dtype": torch.float32},
        {"res": 1024, "total_vram": 23, "bs": 3, "dtype": torch.float32},
        {"res": 512, "total_vram": 23, "bs": 40, "dtype": torch.float16},
        {"res": 768, "total_vram": 23, "bs": 18, "dtype": torch.float16},
        {"res": 1024, "total_vram": 23, "bs": 10, "dtype": torch.float16},
        # tested on GTX1080Ti
        {"res": 512, "total_vram": 10, "bs": 5, "dtype": torch.float32},
        {"res": 768, "total_vram": 10, "bs": 2, "dtype": torch.float32},
        {"res": 512, "total_vram": 10, "bs": 10, "dtype": torch.float16},
        {"res": 768, "total_vram": 10, "bs": 5, "dtype": torch.float16},
        {"res": 1024, "total_vram": 10, "bs": 3, "dtype": torch.float16},
    ]


    if not torch.cuda.is_available():
        return 1

    total_vram = torch.cuda.mem_get_info()[1] / 1024.0**3
    filtered_bs_search_table = [s for s in bs_search_table if s["dtype"] == dtype]
    for settings in sorted(filtered_bs_search_table, key=lambda k: (k["res"], -k["total_vram"])):
        if input_res <= settings["res"] and total_vram >= settings["total_vram"]:
            bs = settings["bs"]
            if bs > ensemble_size:
                bs = ensemble_size
            elif bs > math.ceil(ensemble_size / 2) and bs < ensemble_size: # pylint: disable=chained-comparison
                bs = math.ceil(ensemble_size / 2)
            return bs

    return 1



"""Depth normals representation using SVD."""
from overrides import overrides
import numpy as np

from vre.utils import MemoryData, ReprOut
from vre.vre_video import VREVideo
from vre.representations import ComputeRepresentationMixin
from vre_repository.normals import NormalsRepresentation

from .depth_svd_impl import depths_to_normals

class DepthNormalsSVD(NormalsRepresentation, ComputeRepresentationMixin):
    """
    General method for estimating normals from a depth map (+ intrinsics): a 2D window centered on each pixel is
    projected into 3D and then a plane is fitted on the 3D pointcloud using SVD.
    """
    def __init__(self, sensor_fov: int, sensor_size: tuple[int, int], window_size: int,
                 input_downsample_step: int = None, stride: int = None, **kwargs):
        NormalsRepresentation.__init__(self, **kwargs)
        ComputeRepresentationMixin.__init__(self)
        assert window_size % 2 == 1, f"Expected odd window size. Got: {window_size}"
        self.sensor_fov = sensor_fov
        self.sensor_size = sensor_size
        self.window_size = window_size
        self.stride = stride or 1
        self.input_downsample_step = input_downsample_step or 1
        assert len(self.dependencies) == 1, f"Expected exactly one depth method, got: {self.dependencies}"

    @overrides
    def compute(self, video: VREVideo, ixs: list[int], dep_data: list[ReprOut] | None = None) -> ReprOut:
        assert dep_data is not None and len(dep_data) == 1, dep_data
        assert dep_data[0].key == ixs, (self, dep_data[0], ixs)
        depths = dep_data[0].output
        assert len(depths.shape) == 4 and depths.shape[-1] == 1, f"Expected (B, H, W, 1) got: {depths.shape}"
        res = depths_to_normals(depths[..., 0], self.sensor_fov, self.window_size, self.stride,
                                self.sensor_size, self.input_downsample_step)
        res = (MemoryData(res).astype(np.float32) + 1) / 2
        return ReprOut(frames=video[ixs], output=res, key=ixs)



"""VRE Representation module"""
from __future__ import annotations
from abc import ABC, abstractmethod
import numpy as np

from ..utils import parsed_str_type, ReprOut

class Representation(ABC):
    """Generic Representation class for VRE"""
    def __init__(self, name: str, dependencies: list[Representation] | None = None):
        deps = [] if dependencies is None else dependencies
        assert isinstance(deps, list), type(deps)
        # assert all(isinstance(dep, Representation) for dep in deps), (name, [(dep, type(dep)) for dep in deps])
        self.name = name
        self.dependencies = deps
        # self.data: ReprOut | None = None

    def size(self, repr_out: ReprOut) -> tuple[int, ...]:
        """Returns the (b, h, w, c) tuple of the size of the current representation"""
        return tuple(repr_out.output.shape)

    ## Abstract methods ##
    @abstractmethod
    def make_images(self, data: ReprOut) -> np.ndarray:
        """Given the output of self.compute(video, ixs) of type ReprOut, return a [0:255] image for each frame"""

    @property
    @abstractmethod
    def n_channels(self) -> int:
        """The number of channels (dimensions, not resolution) of this representation"""

    ## Public methods & properties ##
    @property
    def dep_names(self) -> list[str]:
        """return the list of dependencies names"""
        return [r.name for r in self.dependencies]

    @property
    def is_classification(self) -> bool:
        """if we have self.classes. Used in MultiTaskReader."""
        return hasattr(self, "classes") and self.classes is not None # pylint: disable=no-member

    ## Magic methods ##
    def __repr__(self):
        return f"{parsed_str_type(self)}({self.name}{f' {self.dep_names}' if len(self.dep_names) > 0 else ''})"



"""HSV module"""
import numpy as np
from overrides import overrides

from vre.vre_video import VREVideo
from vre.utils import ReprOut
from vre.representations import ComputeRepresentationMixin

from .rgb import RGB
from .color_representation import ColorRepresentation

def rgb2hsv(rgb: np.ndarray) -> np.ndarray:
    """RGB to HSV color space conversion."""
    assert rgb.ndim in (3, 4), f"Expected 3 or 4 dimensions, got {rgb.ndim}"

    arr = rgb.astype(np.float64) / 255 if rgb.dtype == np.uint8 else rgb
    out = np.empty_like(arr)

    # -- V channel
    out_v = arr.max(-1)

    # -- S channel
    delta = np.ptp(arr, axis=-1)
    # Ignore warning for zero divided by zero
    old_settings = np.seterr(invalid='ignore')
    out_s = delta / out_v
    out_s[delta == 0.0] = 0.0

    # -- H channel
    # red is max
    idx = arr[..., 0] == out_v
    out[idx, 0] = (arr[idx, 1] - arr[idx, 2]) / delta[idx]

    # green is max
    idx = arr[..., 1] == out_v
    out[idx, 0] = 2.0 + (arr[idx, 2] - arr[idx, 0]) / delta[idx]

    # blue is max
    idx = arr[..., 2] == out_v
    out[idx, 0] = 4.0 + (arr[idx, 0] - arr[idx, 1]) / delta[idx]
    out_h = (out[..., 0] / 6.0) % 1.0
    out_h[delta == 0.0] = 0.0

    np.seterr(**old_settings)

    # -- output
    out[..., 0] = out_h
    out[..., 1] = out_s
    out[..., 2] = out_v

    # remove NaN
    out[np.isnan(out)] = 0

    return out

class HSV(ColorRepresentation, ComputeRepresentationMixin):
    """HSV representation"""
    def __init__(self, name: str, dependencies: list[str]):
        assert len(dependencies) == 1 and isinstance(dependencies[0], RGB), dependencies
        ColorRepresentation.__init__(self, name, dependencies=dependencies)
        ComputeRepresentationMixin.__init__(self)

    @overrides
    def compute(self, video: VREVideo, ixs: list[int], dep_data: list[ReprOut] | None = None) -> ReprOut:
        assert dep_data is not None and len(dep_data) == 1, dep_data
        return ReprOut(frames=video[ixs], output=rgb2hsv(dep_data[0].output).astype(np.float32), key=ixs)



"""FlowRife representation"""
import numpy as np
import torch as tr
import torch.nn.functional as F
from overrides import overrides

from vre.vre_video import VREVideo
from vre.utils import MemoryData
from vre.representations import ReprOut, LearnedRepresentationMixin, ComputeRepresentationMixin
from vre_repository.optical_flow import OpticalFlowRepresentation
from vre_repository.weights_repository import fetch_weights

from .rife_impl import Model

class FlowRife(OpticalFlowRepresentation, LearnedRepresentationMixin, ComputeRepresentationMixin):
    """FlowRife representation"""
    def __init__(self, compute_backward_flow: bool, uhd: bool, flow_delta_frames: int = 1, **kwargs):
        OpticalFlowRepresentation.__init__(self, **kwargs)
        LearnedRepresentationMixin.__init__(self)
        ComputeRepresentationMixin.__init__(self)
        self.uhd = uhd
        self.flow_delta_frames = flow_delta_frames
        assert compute_backward_flow is False, "Not supported"
        self.no_backward_flow = True if compute_backward_flow is None else not compute_backward_flow
        self.model: Model | None = None

    @overrides
    def compute(self, video: VREVideo, ixs: list[int], dep_data: list[ReprOut] | None = None) -> ReprOut:
        frames = video[ixs]
        right_frames = self.get_delta_frames(video, ixs)
        x_s, x_t, padding = self._preprocess(frames, right_frames)
        with tr.no_grad():
            prediction = self.model.inference(x_s, x_t, self.uhd, self.no_backward_flow)
        flow = self._postprocess(prediction, padding)
        return ReprOut(frames=video[ixs], output=MemoryData(flow), key=ixs)

    @staticmethod
    @overrides
    def weights_repository_links(**kwargs) -> list[str]:
        return [
            "optical_flow/rife/contextnet.ckpt",
            "optical_flow/rife/unet.ckpt",
            "optical_flow/rife/flownet.ckpt"
        ]

    @overrides
    def vre_setup(self, load_weights: bool = True):
        assert self.setup_called is False
        self.model: Model = Model().eval()
        if load_weights:
            paths = fetch_weights(FlowRife.weights_repository_links())
            self.model.load_model(paths[0].parent)
        self.model = self.model.to(self.device)
        self.setup_called = True

    @overrides
    def vre_free(self):
        assert self.setup_called is True and self.model is not None, (self.setup_called, self.model is not None)
        if str(self.device).startswith("cuda"):
            self.model.to("cpu")
            tr.cuda.empty_cache()
        self.model = None
        self.setup_called = False

    def _preprocess(self, sources: np.ndarray, targets: np.ndarray) -> tuple[tr.Tensor, tr.Tensor, tuple]:
        # Convert, preprocess & pad
        sources = sources.transpose(0, 3, 1, 2)
        targets = targets.transpose(0, 3, 1, 2)
        tr_sources = tr.from_numpy(sources).to(self.device).float() / 255.0
        tr_targets = tr.from_numpy(targets).to(self.device).float() / 255.0
        h, w = tr_sources.shape[2:4]
        ph = ((h - 1) // 32 + 1) * 32
        pw = ((w - 1) // 32 + 1) * 32
        padding = (0, pw - w, 0, ph - h)
        tr_sources_padded = F.pad(tr_sources, padding) # pylint: disable=not-callable
        tr_target_padded = F.pad(tr_targets, padding) # pylint: disable=not-callable
        return tr_sources_padded, tr_target_padded, padding

    def _postprocess(self, prediction: tr.Tensor, padding: tuple) -> np.ndarray:
        flow = prediction.cpu().numpy().transpose(0, 2, 3, 1) # (B, H, W, C)
        returned_shape = flow.shape[1:3]
        # Remove the padding to keep original shape
        half_ph, half_pw = padding[3] // 2, padding[1] // 2
        flow = flow[:, 0: returned_shape[0] - half_ph, 0: returned_shape[1] - half_pw]
        flow = flow / returned_shape # [-px : px] => [-1 : 1]
        return flow.astype(np.float32)



# pylint: disable=all
from typing import List, Tuple
import torch
from torch import device

class Boxes:
    """
    This structure stores a list of boxes as a Nx4 torch.Tensor.
    It supports some common methods about boxes
    (`area`, `clip`, `nonempty`, etc),
    and also behaves like a Tensor
    (support indexing, `to(device)`, `.device`, and iteration over all boxes)

    Attributes:
        tensor (torch.Tensor): float matrix of Nx4. Each row is (x1, y1, x2, y2).
    """

    def __init__(self, tensor: torch.Tensor):
        """
        Args:
            tensor (Tensor[float]): a Nx4 matrix.  Each row is (x1, y1, x2, y2).
        """
        if not isinstance(tensor, torch.Tensor):
            tensor = torch.as_tensor(tensor, dtype=torch.float32, device=torch.device("cpu"))
        else:
            tensor = tensor.to(torch.float32)
        if tensor.numel() == 0:
            # Use reshape, so we don't end up creating a new tensor that does not depend on
            # the inputs (and consequently confuses jit)
            tensor = tensor.reshape((-1, 4)).to(dtype=torch.float32)
        assert tensor.dim() == 2 and tensor.size(-1) == 4, tensor.size()

        self.tensor = tensor

    def clone(self) -> "Boxes":
        """
        Clone the Boxes.

        Returns:
            Boxes
        """
        return Boxes(self.tensor.clone())

    def to(self, device: torch.device):
        # Boxes are assumed float32 and does not support to(dtype)
        return Boxes(self.tensor.to(device=device))

    def area(self) -> torch.Tensor:
        """
        Computes the area of all the boxes.

        Returns:
            torch.Tensor: a vector with areas of each box.
        """
        box = self.tensor
        area = (box[:, 2] - box[:, 0]) * (box[:, 3] - box[:, 1])
        return area

    def clip(self, box_size: Tuple[int, int]) -> None:
        """
        Clip (in place) the boxes by limiting x coordinates to the range [0, width]
        and y coordinates to the range [0, height].

        Args:
            box_size (height, width): The clipping box's size.
        """
        assert torch.isfinite(self.tensor).all(), "Box tensor contains infinite or NaN!"
        h, w = box_size
        x1 = self.tensor[:, 0].clamp(min=0, max=w)
        y1 = self.tensor[:, 1].clamp(min=0, max=h)
        x2 = self.tensor[:, 2].clamp(min=0, max=w)
        y2 = self.tensor[:, 3].clamp(min=0, max=h)
        self.tensor = torch.stack((x1, y1, x2, y2), dim=-1)

    def nonempty(self, threshold: float = 0.0) -> torch.Tensor:
        """
        Find boxes that are non-empty.
        A box is considered empty, if either of its side is no larger than threshold.

        Returns:
            Tensor:
                a binary vector which represents whether each box is empty
                (False) or non-empty (True).
        """
        box = self.tensor
        widths = box[:, 2] - box[:, 0]
        heights = box[:, 3] - box[:, 1]
        keep = (widths > threshold) & (heights > threshold)
        return keep

    def __getitem__(self, item) -> "Boxes":
        """
        Args:
            item: int, slice, or a BoolTensor

        Returns:
            Boxes: Create a new :class:`Boxes` by indexing.

        The following usage are allowed:

        1. `new_boxes = boxes[3]`: return a `Boxes` which contains only one box.
        2. `new_boxes = boxes[2:10]`: return a slice of boxes.
        3. `new_boxes = boxes[vector]`, where vector is a torch.BoolTensor
           with `length = len(boxes)`. Nonzero elements in the vector will be selected.

        Note that the returned Boxes might share storage with this Boxes,
        subject to Pytorch's indexing semantics.
        """
        if isinstance(item, int):
            return Boxes(self.tensor[item].view(1, -1))
        b = self.tensor[item]
        assert b.dim() == 2, "Indexing on Boxes with {} failed to return a matrix!".format(item)
        return Boxes(b)

    def __len__(self) -> int:
        return self.tensor.shape[0]

    def __repr__(self) -> str:
        return "Boxes(" + str(self.tensor) + ")"

    def inside_box(self, box_size: Tuple[int, int], boundary_threshold: int = 0) -> torch.Tensor:
        """
        Args:
            box_size (height, width): Size of the reference box.
            boundary_threshold (int): Boxes that extend beyond the reference box
                boundary by more than boundary_threshold are considered "outside".

        Returns:
            a binary vector, indicating whether each box is inside the reference box.
        """
        height, width = box_size
        inds_inside = (
            (self.tensor[..., 0] >= -boundary_threshold)
            & (self.tensor[..., 1] >= -boundary_threshold)
            & (self.tensor[..., 2] < width + boundary_threshold)
            & (self.tensor[..., 3] < height + boundary_threshold)
        )
        return inds_inside

    def get_centers(self) -> torch.Tensor:
        """
        Returns:
            The box centers in a Nx2 array of (x, y).
        """
        return (self.tensor[:, :2] + self.tensor[:, 2:]) / 2

    def scale(self, scale_x: float, scale_y: float) -> None:
        """
        Scale the box with horizontal and vertical scaling factors
        """
        self.tensor[:, 0::2] *= scale_x
        self.tensor[:, 1::2] *= scale_y

    @classmethod
    def cat(cls, boxes_list: List["Boxes"]) -> "Boxes":
        """
        Concatenates a list of Boxes into a single Boxes

        Arguments:
            boxes_list (list[Boxes])

        Returns:
            Boxes: the concatenated Boxes
        """
        assert isinstance(boxes_list, (list, tuple))
        if len(boxes_list) == 0:
            return cls(torch.empty(0))
        assert all([isinstance(box, Boxes) for box in boxes_list])

        # use torch.cat (v.s. layers.cat) so the returned boxes never share storage with input
        cat_boxes = cls(torch.cat([b.tensor for b in boxes_list], dim=0))
        return cat_boxes

    @property
    def device(self) -> device:
        return self.tensor.device

    # type "Iterator[torch.Tensor]", yield, and iter() not supported by torchscript
    # https://github.com/pytorch/pytorch/issues/18627
    @torch.jit.unused
    def __iter__(self):
        """
        Yield a box as a Tensor of shape (4,) at a time.
        """
        yield from self.tensor



"""optical_flow_representation.py -- module implementing an Optical Flow Represenatation generic class"""
from overrides import overrides
import numpy as np
from vre.utils import colorize_optical_flow, ReprOut, clip
from vre.representations import Representation, NpIORepresentation, NormedRepresentationMixin
from vre.vre_video import VREVideo

def _get_delta_frames(video: VREVideo, ixs: list[int], delta: int) -> list[int]:
    assert delta != 0, delta
    return [clip(ix + delta, 0, len(video) - 1) for ix in ixs]

class OpticalFlowRepresentation(Representation, NpIORepresentation, NormedRepresentationMixin):
    """OpticalFlowRepresentation. Implements flow task-specific stuff."""
    def __init__(self, name: str, delta: int, **kwargs):
        Representation.__init__(self, name, **kwargs)
        NpIORepresentation.__init__(self)
        NormedRepresentationMixin.__init__(self)
        assert isinstance(delta, int) and delta != 0, delta
        self.delta = delta

    @property
    @overrides
    def n_channels(self) -> int:
        return 2

    @overrides
    def make_images(self, data: ReprOut) -> np.ndarray:
        y = self.unnormalize(data.output) if self.normalization is not None else data.output
        return colorize_optical_flow(y)

    def get_delta_frames(self, video: VREVideo, ixs: list[int]) -> np.ndarray:
        """for a given list of frames at .compute() time, return the delta frames required to compute the flow"""
        return video[_get_delta_frames(video, ixs, self.delta)]



"""colorize_depth -- colorized depth maps"""
import numpy as np

def colorize_depth(depth_map: np.ndarray, min_max_depth: tuple[float, float] | None = None,
                   percentiles: tuple[int, int] | None = None) -> np.ndarray:
    """Colorize depth maps. Returns a [0:1] (H, W, 3) float image. Batched."""
    assert len(depth_map.shape) == 4 and depth_map.shape[-1] == 1, depth_map.shape
    assert isinstance(depth_map, np.ndarray), depth_map
    assert (min_max_depth is not None and percentiles is None) or min_max_depth is None, (min_max_depth, percentiles)
    dm_no_nan = np.nan_to_num(depth_map[..., 0], True, 0, 0, 0)
    percentiles = [0, 100] if percentiles is None else percentiles
    min_depth, max_depth = np.percentile(dm_no_nan, percentiles)
    if min_depth == max_depth:
        return (dm_no_nan[..., None] * 0).repeat(3, axis=-1)
    dm_min_max = (dm_no_nan - min_depth) / (max_depth - min_depth)
    return _spectral(dm_min_max.clip(0, 1))

def _spectral(grayscale_array: np.ndarray):
    """
    Maps a grayscale array (values between 0 and 1) to the spectral colormap without using Matplotlib.
    Args:
        grayscale_array (np.ndarray): Input array with values in the range [0, 1].
    Returns:
        np.ndarray: RGB array with values in the range [0, 1], shape (H, W, 3).
    """
    assert np.issubsctype(grayscale_array, np.floating), grayscale_array.dtype
    assert (_min := grayscale_array.min()) >= 0 and grayscale_array.max() <= 1, (_min, grayscale_array.max())
    spectral_colors = np.array([
        [158/255,   1/255,  66/255],  # Dark Red
        [213/255,  62/255,  79/255],  # Red
        [244/255, 109/255,  67/255],  # Orange
        [253/255, 174/255,  97/255],  # Light Orange
        [254/255, 224/255, 139/255],  # Yellow
        [230/255, 245/255, 152/255],  # Light Green
        [171/255, 221/255, 164/255],  # Green
        [102/255, 194/255, 165/255],  # Cyan
        [ 50/255, 136/255, 189/255],  # Blue
        [ 94/255,  79/255, 162/255]   # Purple
    ])

    # Corresponding grayscale positions for the colormap
    spectral_positions = np.linspace(0, 1, len(spectral_colors))

    # Interpolate each color channel separately
    red = np.interp(grayscale_array, spectral_positions, spectral_colors[:, 0])
    green = np.interp(grayscale_array, spectral_positions, spectral_colors[:, 1])
    blue = np.interp(grayscale_array, spectral_positions, spectral_colors[:, 2])

    # Stack the interpolated channels to form the RGB array
    rgb_array = np.stack([red, green, blue], axis=-1)
    return rgb_array.astype(np.float32)



# pylint: disable=all
# Data loading based on https://github.com/NVIDIA/flownet2-pytorch

import numpy as np
import torch
import torch.utils.data as data
import torch.nn.functional as F

import os
import math
import random
from glob import glob
import os.path as osp

from utils import frame_utils
from utils.augmentor import FlowAugmentor, SparseFlowAugmentor


class FlowDataset(data.Dataset):
    def __init__(self, aug_params=None, sparse=False):
        self.augmentor = None
        self.sparse = sparse
        if aug_params is not None:
            if sparse:
                self.augmentor = SparseFlowAugmentor(**aug_params)
            else:
                self.augmentor = FlowAugmentor(**aug_params)

        self.is_test = False
        self.init_seed = False
        self.flow_list = []
        self.image_list = []
        self.extra_info = []

    def __getitem__(self, index):

        if self.is_test:
            img1 = frame_utils.read_gen(self.image_list[index][0])
            img2 = frame_utils.read_gen(self.image_list[index][1])
            img1 = np.array(img1).astype(np.uint8)[..., :3]
            img2 = np.array(img2).astype(np.uint8)[..., :3]
            img1 = torch.from_numpy(img1).permute(2, 0, 1).float()
            img2 = torch.from_numpy(img2).permute(2, 0, 1).float()
            return img1, img2, self.extra_info[index]

        if not self.init_seed:
            worker_info = torch.utils.data.get_worker_info()
            if worker_info is not None:
                torch.manual_seed(worker_info.id)
                np.random.seed(worker_info.id)
                random.seed(worker_info.id)
                self.init_seed = True

        index = index % len(self.image_list)
        valid = None
        if self.sparse:
            flow, valid = frame_utils.readFlowKITTI(self.flow_list[index])
        else:
            flow = frame_utils.read_gen(self.flow_list[index])

        img1 = frame_utils.read_gen(self.image_list[index][0])
        img2 = frame_utils.read_gen(self.image_list[index][1])

        flow = np.array(flow).astype(np.float32)
        img1 = np.array(img1).astype(np.uint8)
        img2 = np.array(img2).astype(np.uint8)

        # grayscale images
        if len(img1.shape) == 2:
            img1 = np.tile(img1[...,None], (1, 1, 3))
            img2 = np.tile(img2[...,None], (1, 1, 3))
        else:
            img1 = img1[..., :3]
            img2 = img2[..., :3]

        if self.augmentor is not None:
            if self.sparse:
                img1, img2, flow, valid = self.augmentor(img1, img2, flow, valid)
            else:
                img1, img2, flow = self.augmentor(img1, img2, flow)

        img1 = torch.from_numpy(img1).permute(2, 0, 1).float()
        img2 = torch.from_numpy(img2).permute(2, 0, 1).float()
        flow = torch.from_numpy(flow).permute(2, 0, 1).float()

        if valid is not None:
            valid = torch.from_numpy(valid)
        else:
            valid = (flow[0].abs() < 1000) & (flow[1].abs() < 1000)

        return img1, img2, flow, valid.float()


    def __rmul__(self, v):
        self.flow_list = v * self.flow_list
        self.image_list = v * self.image_list
        return self

    def __len__(self):
        return len(self.image_list)


class MpiSintel(FlowDataset):
    def __init__(self, aug_params=None, split='training', root='datasets/Sintel', dstype='clean'):
        super(MpiSintel, self).__init__(aug_params)
        flow_root = osp.join(root, split, 'flow')
        image_root = osp.join(root, split, dstype)

        if split == 'test':
            self.is_test = True

        for scene in os.listdir(image_root):
            image_list = sorted(glob(osp.join(image_root, scene, '*.png')))
            for i in range(len(image_list)-1):
                self.image_list += [ [image_list[i], image_list[i+1]] ]
                self.extra_info += [ (scene, i) ] # scene and frame_id

            if split != 'test':
                self.flow_list += sorted(glob(osp.join(flow_root, scene, '*.flo')))


class FlyingChairs(FlowDataset):
    def __init__(self, aug_params=None, split='train', root='datasets/FlyingChairs_release/data'):
        super(FlyingChairs, self).__init__(aug_params)

        images = sorted(glob(osp.join(root, '*.ppm')))
        flows = sorted(glob(osp.join(root, '*.flo')))
        assert (len(images)//2 == len(flows))

        split_list = np.loadtxt('chairs_split.txt', dtype=np.int32)
        for i in range(len(flows)):
            xid = split_list[i]
            if (split=='training' and xid==1) or (split=='validation' and xid==2):
                self.flow_list += [ flows[i] ]
                self.image_list += [ [images[2*i], images[2*i+1]] ]


class FlyingThings3D(FlowDataset):
    def __init__(self, aug_params=None, root='datasets/FlyingThings3D', dstype='frames_cleanpass'):
        super(FlyingThings3D, self).__init__(aug_params)

        for cam in ['left']:
            for direction in ['into_future', 'into_past']:
                image_dirs = sorted(glob(osp.join(root, dstype, 'TRAIN/*/*')))
                image_dirs = sorted([osp.join(f, cam) for f in image_dirs])

                flow_dirs = sorted(glob(osp.join(root, 'optical_flow/TRAIN/*/*')))
                flow_dirs = sorted([osp.join(f, direction, cam) for f in flow_dirs])

                for idir, fdir in zip(image_dirs, flow_dirs):
                    images = sorted(glob(osp.join(idir, '*.png')) )
                    flows = sorted(glob(osp.join(fdir, '*.pfm')) )
                    for i in range(len(flows)-1):
                        if direction == 'into_future':
                            self.image_list += [ [images[i], images[i+1]] ]
                            self.flow_list += [ flows[i] ]
                        elif direction == 'into_past':
                            self.image_list += [ [images[i+1], images[i]] ]
                            self.flow_list += [ flows[i+1] ]
      

class KITTI(FlowDataset):
    def __init__(self, aug_params=None, split='training', root='datasets/KITTI'):
        super(KITTI, self).__init__(aug_params, sparse=True)
        if split == 'testing':
            self.is_test = True

        root = osp.join(root, split)
        images1 = sorted(glob(osp.join(root, 'image_2/*_10.png')))
        images2 = sorted(glob(osp.join(root, 'image_2/*_11.png')))

        for img1, img2 in zip(images1, images2):
            frame_id = img1.split('/')[-1]
            self.extra_info += [ [frame_id] ]
            self.image_list += [ [img1, img2] ]

        if split == 'training':
            self.flow_list = sorted(glob(osp.join(root, 'flow_occ/*_10.png')))


class HD1K(FlowDataset):
    def __init__(self, aug_params=None, root='datasets/HD1k'):
        super(HD1K, self).__init__(aug_params, sparse=True)

        seq_ix = 0
        while 1:
            flows = sorted(glob(os.path.join(root, 'hd1k_flow_gt', 'flow_occ/%06d_*.png' % seq_ix)))
            images = sorted(glob(os.path.join(root, 'hd1k_input', 'image_2/%06d_*.png' % seq_ix)))

            if len(flows) == 0:
                break

            for i in range(len(flows)-1):
                self.flow_list += [flows[i]]
                self.image_list += [ [images[i], images[i+1]] ]

            seq_ix += 1


def fetch_dataloader(args, TRAIN_DS='C+T+K+S+H'):
    """ Create the data loader for the corresponding trainign set """

    if args.stage == 'chairs':
        aug_params = {'crop_size': args.image_size, 'min_scale': -0.1, 'max_scale': 1.0, 'do_flip': True}
        train_dataset = FlyingChairs(aug_params, split='training')
    
    elif args.stage == 'things':
        aug_params = {'crop_size': args.image_size, 'min_scale': -0.4, 'max_scale': 0.8, 'do_flip': True}
        clean_dataset = FlyingThings3D(aug_params, dstype='frames_cleanpass')
        final_dataset = FlyingThings3D(aug_params, dstype='frames_finalpass')
        train_dataset = clean_dataset + final_dataset

    elif args.stage == 'sintel':
        aug_params = {'crop_size': args.image_size, 'min_scale': -0.2, 'max_scale': 0.6, 'do_flip': True}
        things = FlyingThings3D(aug_params, dstype='frames_cleanpass')
        sintel_clean = MpiSintel(aug_params, split='training', dstype='clean')
        sintel_final = MpiSintel(aug_params, split='training', dstype='final')        

        if TRAIN_DS == 'C+T+K+S+H':
            kitti = KITTI({'crop_size': args.image_size, 'min_scale': -0.3, 'max_scale': 0.5, 'do_flip': True})
            hd1k = HD1K({'crop_size': args.image_size, 'min_scale': -0.5, 'max_scale': 0.2, 'do_flip': True})
            train_dataset = 100*sintel_clean + 100*sintel_final + 200*kitti + 5*hd1k + things

        elif TRAIN_DS == 'C+T+K/S':
            train_dataset = 100*sintel_clean + 100*sintel_final + things

    elif args.stage == 'kitti':
        aug_params = {'crop_size': args.image_size, 'min_scale': -0.2, 'max_scale': 0.4, 'do_flip': False}
        train_dataset = KITTI(aug_params, split='training')

    train_loader = data.DataLoader(train_dataset, batch_size=args.batch_size, 
        pin_memory=False, shuffle=True, num_workers=4, drop_last=True)

    print('Training with %d image pairs' % len(train_dataset))
    return train_loader




"""FastSAM representation."""
from argparse import ArgumentParser, Namespace
from datetime import datetime
from pathlib import Path
from overrides import overrides
import numpy as np
import torch as tr
from torch import nn
from torch.nn import functional as F

from vre.vre_video import VREVideo, FakeVideo
from vre.representations import (Representation, ReprOut, LearnedRepresentationMixin,
                                 ComputeRepresentationMixin, NpIORepresentation, NormedRepresentationMixin)
from vre.utils import image_resize_batch, image_read, image_write, MemoryData
from vre.logger import vre_logger as logger
from vre_repository.weights_repository import fetch_weights

try:
    from .fastsam_impl import (
        FastSAM as Model, FastSAMPredictor, FastSAMPrompt, Results, bbox_iou, non_max_suppression, process_mask_native)
except ImportError:
    from vre_repository.soft_segmentation.fastsam.fastsam_impl import (
        FastSAM as Model, FastSAMPredictor, FastSAMPrompt, Results, bbox_iou, non_max_suppression, process_mask_native)

class FastSam(Representation, LearnedRepresentationMixin, ComputeRepresentationMixin,
              NpIORepresentation, NormedRepresentationMixin):
    """FastSAM representation."""
    def __init__(self, variant: str, iou: float, conf: float, **kwargs):
        Representation.__init__(self, **kwargs)
        LearnedRepresentationMixin.__init__(self)
        ComputeRepresentationMixin.__init__(self)
        NpIORepresentation.__init__(self)
        NormedRepresentationMixin.__init__(self)
        assert variant in ("fastsam-s", "fastsam-x", "testing"), variant
        self.variant = variant
        self.conf = conf
        self.iou = iou
        self.model: nn.Module | None = None
        self._imgsz: int | None = None

    @overrides
    def compute(self, video: VREVideo, ixs: list[int], dep_data: list[ReprOut] | None = None) -> ReprOut:
        tr_x = self._preproces(frames := video[ixs])
        tr_y: tr.FloatTensor = self.model.forward(tr_x)
        mb, _, i_h, i_w = tr_x.shape[0:4]
        boxes = self._postprocess(preds=tr_y, inference_height=i_h, inference_width=i_w, conf=self.conf, iou=self.iou)
        # Note: image_size' see the commnet in resize()
        extra = [{"boxes": boxes[i].to("cpu").numpy(), "inference_size": (i_h, i_w), "image_size": frames.shape[1:3]}
                 for i in range(mb)]
        assert len(tr_y[1]) == 3, len(tr_y[1])
        # Note: this is called 'proto' in the original implementation. Only this part of predictions is used for plots.
        return ReprOut(frames=frames, output=MemoryData(tr_y[1][-1].to("cpu").numpy()), extra=extra, key=ixs)

    @overrides(check_signature=False)
    def make_images(self, data: ReprOut) -> np.ndarray:
        y_fastsam, extra = data.output, data.extra # len(y_fastsam) == len(extra) == len(ixs)
        assert all((inf_size := e["inference_size"]) == extra[0]["inference_size"] for e in extra), extra
        assert all((img_size := e["image_size"]) == extra[0]["image_size"] for e in extra), extra
        tr_y = tr.from_numpy(y_fastsam).to(self.device)
        boxes = [tr.from_numpy(e["boxes"]).to(self.device) for e in extra]
        # scaled_boxes = [scale_box(box, *inf_size, *img_size) for box in boxes] # TODO: needed for faster proc?

        res: list[np.ndarray] = []
        for i, (box, frame) in enumerate(zip(boxes, data.frames)):
            if len(box) == 0:  # save empty boxes
                res.append(frame)
                continue
            masks = process_mask_native(tr_y[i], box[:, 6:], box[:, :4], inf_size)
            res_i = Results(orig_img=frame, path=None, names={0: "object"}, boxes=box[:, 0:6], masks=masks)
            prompt_process = FastSAMPrompt(frame, [res_i], device=self.device)
            ann = prompt_process.results[0].masks.data
            prompt_res = prompt_process.plot_to_result(annotations=ann, withContours=True)
            res.append(prompt_res)
        res_arr = image_resize_batch(res, *img_size, interpolation="bilinear")
        return res_arr

    @overrides
    def size(self, repr_out: ReprOut) -> tuple[int, ...]:
        return (len(repr_out.output), *repr_out.extra[0]["image_size"], 3) # Note: not the embeddings size!

    @overrides
    def resize(self, data: ReprOut, new_size: tuple[int, int]) -> ReprOut:
        assert data is not None
        # Note: check the old implementation. We used to rescale the boxes, now we just keep them as-is and
        # only resize the final output. It's debatable if we want to resize the boxes or not to maintain quality.
        new_extra = [{**e, "image_size": new_size} for e in data.extra]
        output_images = None
        if data.output_images is not None:
            output_images = image_resize_batch(data.output_images, *new_size, interpolation="nearest")
        return ReprOut(frames=data.frames, output=data.output, extra=new_extra,
                       key=data.key, output_images=output_images)

    @staticmethod
    @overrides
    def weights_repository_links(**kwargs) -> list[str]:
        match kwargs["variant"]:
            case "fastsam-x": return ["soft_segmentation/fastsam/FastSAM-x.pt"]
            case "fastsam-s": return ["soft_segmentation/fastsam/FastSAM-s.pt"]
            case "testing": return []
            case _: raise NotImplementedError(kwargs)

    @overrides
    def vre_setup(self, load_weights: bool = True):
        assert self.setup_called is False
        assert load_weights is False or (load_weights is True and self.variant != "testing"), "no weights for testing"
        weights_file = "testing"
        if self.variant != "testing":
            weights_file = fetch_weights(FastSam.weights_repository_links(variant=self.variant))[0]
        self._imgsz = 1024 if self.variant != "testing" else 64
        _overrides = {"task": "segment", "imgsz": self._imgsz, "single_cls": False, "model": str(weights_file),
                      "conf": self.conf, "device": "cpu", "retina_masks": False, "iou": self.iou,
                      "mode": "predict", "save": False}
        predictor = FastSAMPredictor(overrides=_overrides)
        _model = Model(str(weights_file))
        predictor.setup_model(model=_model.model, verbose=False)
        self.model = predictor.model.eval().to(self.device)
        self.setup_called = True

    @overrides
    def vre_free(self):
        assert self.setup_called is True and self.model is not None, (self.setup_called, self.model is not None)
        if str(self.device).startswith("cuda"):
            self.model.to("cpu")
            tr.cuda.empty_cache()
        self.model = None
        self.setup_called = False

    @property
    @overrides
    def n_channels(self) -> int:
        raise ValueError("I hate inheritance. Makes no sense for this representation")

    def _postprocess(self, preds: tr.Tensor, inference_height: int, inference_width: int,
                     conf: float, iou: float) -> list[tr.Tensor]:
        p = non_max_suppression(preds[0], conf, iou, agnostic=False, max_det=300, nc=1, classes=None)

        for _p in p:
            if len(_p) == 0:
                continue
            full_box = tr.zeros_like(_p[0])
            full_box[2], full_box[3], full_box[4], full_box[6:] = inference_width, inference_height, 1.0, 1.0
            full_box = full_box.view(1, -1)
            critical_iou_index = bbox_iou(full_box[0][:4], _p[:, :4], iou_thres=0.9,
                                          image_shape=(inference_height, inference_width))
            if critical_iou_index.numel() != 0:
                full_box[0][4] = _p[critical_iou_index][:, 4]
                full_box[0][6:] = _p[critical_iou_index][:, 6:]
                _p[critical_iou_index] = full_box
        return p

    def _preproces(self, x: np.ndarray) -> tr.Tensor:
        x = x.astype(np.float32) / 255
        desired_max = self._imgsz
        tr_x = tr.from_numpy(x).permute(0, 3, 1, 2).to(self.device)
        w, h = x.shape[1:3]
        if w > h:
            new_w, new_h = h * desired_max // w, desired_max
        else:
            new_w, new_h = desired_max, w * desired_max // h
        def _offset32(x: int) -> int:
            return ((32 - x % 32) if x < 32 or (x % 32 <= 16) else -(x % 32)) % 32 # get closest modulo 32 offset of x
        new_h, new_w = new_h + _offset32(new_h), new_w + _offset32(new_w) # needed because it throws otherwise
        tr_x: tr.FloatTensor = F.interpolate(tr_x, size=(new_h, new_w), mode="bilinear", align_corners=False)
        return tr_x.to(self.device)

def get_args() -> Namespace:
    """cli args"""
    parser = ArgumentParser()
    parser = ArgumentParser()
    parser.add_argument("model_id", choices=["fastsam-x", "fastsam-s", "testing"])
    parser.add_argument("input_image", type=Path)
    parser.add_argument("output_path", type=Path)
    return parser.parse_args()

def main(args: Namespace):
    """main fn. Usage: python fastsam.py fastsam-x/fastsam-s demo1.jpg output1.jpg"""
    img = image_read(args.input_image)

    fastsam = FastSam(args.model_id, iou=0.9, conf=0.4, name="fastsam", dependencies=[])
    fastsam.device = "cuda" if tr.cuda.is_available() else "cpu"
    fastsam.vre_setup(load_weights=args.model_id != "testing")
    now = datetime.now()
    pred = fastsam.compute(FakeVideo(img[None], 1), [0])
    logger.info(f"Pred took: {datetime.now() - now}")
    semantic_result = fastsam.make_images(pred)[0]
    image_write(semantic_result, args.output_path)
    logger.info(f"Written result to '{args.output_path}'")

    output_path_rsz = args.output_path.parent / f"{args.output_path.stem}_rsz{args.output_path.suffix}"
    pred_rsz = fastsam.resize(pred, (300, 1024))
    semantic_result_rsz = fastsam.make_images(pred_rsz)[0]
    image_write(semantic_result_rsz, output_path_rsz)
    logger.info(f"Written resized result to '{output_path_rsz}'")

    rtol = 1e-2 if tr.cuda.is_available() else 1e-5
    if args.input_image.name == "demo1.jpg" and args.model_id == "fastsam-s":
        assert np.allclose(a := pred.output.mean(), 0.8813972, rtol=rtol), a
        assert np.allclose(b := pred.extra[0]["boxes"].mean(), 46.200043, rtol=rtol), b
        # Note: now only the tuple size and make_images is changed. To be decided if it's worth to downscale preds.
        assert np.allclose(a := pred_rsz.output.mean(), 0.8813972, rtol=rtol), a
        assert np.allclose(b := pred_rsz.extra[0]["boxes"].mean(), 46.200043, rtol=rtol), b
        logger.info("Mean and std test passed!")
    elif args.input_image.name == "demo1.jpg" and args.model_id == "fastsam-x":
        assert np.allclose(a := pred.output.mean(), 0.80122584, rtol=rtol), a
        assert np.allclose(b := pred.extra[0]["boxes"].mean(), 49.640644, rtol=rtol), b
        assert np.allclose(a := pred_rsz.output.mean(), 0.80122584, rtol=rtol), a
        assert np.allclose(b := pred_rsz.extra[0]["boxes"].mean(), 49.640644, rtol=rtol), b
        logger.info("Mean and std test passed!")

if __name__ == "__main__":
    main(get_args())



# Ultralytics YOLO 🚀, AGPL-3.0 license
"""
Ultralytics modules. Visualize with:

from ultralytics.nn.modules import *
import torch
import os

x = torch.ones(1, 128, 40, 40)
m = Conv(128, 128)
f = f'{m._get_name()}.onnx'
torch.onnx.export(m, x, f)
os.system(f'onnxsim {f} {f} && open {f}')
"""

from .block import (C1, C2, C3, C3TR, DFL, SPP, SPPF, Bottleneck, BottleneckCSP, C2f, C3Ghost, C3x, GhostBottleneck,
                    HGBlock, HGStem, Proto, RepC3)
from .conv import (CBAM, ChannelAttention, Concat, Conv, Conv2, ConvTranspose, DWConv, DWConvTranspose2d, Focus,
                   GhostConv, LightConv, RepConv, SpatialAttention)
from .head import Classify, Detect, Pose, RTDETRDecoder, Segment
from .transformer import (AIFI, MLP, DeformableTransformerDecoder, DeformableTransformerDecoderLayer, LayerNorm2d,
                          MLPBlock, MSDeformAttn, TransformerBlock, TransformerEncoderLayer, TransformerLayer)

__all__ = ('Conv', 'Conv2', 'LightConv', 'RepConv', 'DWConv', 'DWConvTranspose2d', 'ConvTranspose', 'Focus',
           'GhostConv', 'ChannelAttention', 'SpatialAttention', 'CBAM', 'Concat', 'TransformerLayer',
           'TransformerBlock', 'MLPBlock', 'LayerNorm2d', 'DFL', 'HGBlock', 'HGStem', 'SPP', 'SPPF', 'C1', 'C2', 'C3',
           'C2f', 'C3x', 'C3TR', 'C3Ghost', 'GhostBottleneck', 'Bottleneck', 'BottleneckCSP', 'Proto', 'Detect',
           'Segment', 'Pose', 'Classify', 'TransformerEncoderLayer', 'RepC3', 'RTDETRDecoder', 'AIFI',
           'DeformableTransformerDecoder', 'DeformableTransformerDecoderLayer', 'MSDeformAttn', 'MLP')



"""DPT Depth Estimation representation"""
import numpy as np
import torch as tr
import torch.nn.functional as F
from overrides import overrides

from vre.vre_video import VREVideo
from vre.utils import MemoryData
from vre.representations import ReprOut, LearnedRepresentationMixin, ComputeRepresentationMixin
from vre_repository.weights_repository import fetch_weights
from vre_repository.depth import DepthRepresentation

from .dpt_impl import DPTDepthModel, get_size

class DepthDpt(DepthRepresentation, LearnedRepresentationMixin, ComputeRepresentationMixin):
    """DPT Depth Estimation representation"""
    def __init__(self, **kwargs):
        LearnedRepresentationMixin.__init__(self)
        ComputeRepresentationMixin.__init__(self)
        DepthRepresentation.__init__(self, min_depth=0, max_depth=1, **kwargs)
        self.net_w, self.net_h = 384, 384
        self.multiple_of = 32
        tr.manual_seed(42)
        self.model: DPTDepthModel | None = None

    @overrides
    def compute(self, video: VREVideo, ixs: list[int], dep_data: list[ReprOut] | None = None) -> ReprOut:
        tr_frames = self._preprocess(video[ixs])
        with tr.no_grad():
            predictions = self.model(tr_frames)
        res = self._postprocess(predictions)
        return ReprOut(frames=video[ixs], output=MemoryData(res), key=ixs)

    @staticmethod
    @overrides
    def weights_repository_links(**kwargs) -> list[str]:
        return ["depth/dpt/depth_dpt_midas.pth"]

    @overrides
    def vre_setup(self, load_weights: bool = True):
        assert self.setup_called is False
        self.model = DPTDepthModel(backbone="vitl16_384", non_negative=True).to("cpu")
        if load_weights:
            path = fetch_weights(DepthDpt.weights_repository_links())[0]
            self.model.load_state_dict(tr.load(path, map_location="cpu"))
        self.model = self.model.eval().to(self.device)
        self.setup_called = True

    @overrides
    def vre_free(self):
        assert self.setup_called is True and self.model is not None, (self.setup_called, self.model is not None)
        if str(self.device).startswith("cuda"):
            self.model.to("cpu")
            tr.cuda.empty_cache()
        self.model = None
        self.setup_called = False

    def _preprocess(self, x: np.ndarray) -> tr.Tensor:
        tr_frames = tr.from_numpy(x).to(self.device)
        tr_frames_perm = tr_frames.permute(0, 3, 1, 2).float() / 255
        curr_h, curr_w = tr_frames.shape[1], tr_frames.shape[2]
        h, w = get_size(self.net_h, self.net_w, curr_h, curr_w, multiple_of=self.multiple_of)
        tr_frames_resized = F.interpolate(tr_frames_perm, size=(h, w), mode="bicubic")
        tr_frames_norm = (tr_frames_resized - 0.5) / 0.5
        return tr_frames_norm

    def _postprocess(self, y: tr.Tensor) -> np.ndarray:
        return (1 / y).clip(0, 1).cpu().numpy()[..., None]



# pylint: disable=all
import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F

try:
    from .warplayer import warp
except ImportError:
    from warplayer import warp

def conv_wo_act(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=True),
    )


def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=True),
        nn.PReLU(out_planes)
    )

class IFBlock(nn.Module):
    def __init__(self, in_planes, scale=1, c=64):
        super(IFBlock, self).__init__()
        self.scale = scale
        self.conv0 = nn.Sequential(
            conv(in_planes, c, 3, 2, 1),
            conv(c, 2*c, 3, 2, 1),
            )
        self.convblock = nn.Sequential(
            conv(2*c, 2*c),
            conv(2*c, 2*c),
            conv(2*c, 2*c),
            conv(2*c, 2*c),
            conv(2*c, 2*c),
            conv(2*c, 2*c),
        )
        self.conv1 = nn.ConvTranspose2d(2*c, 4, 4, 2, 1)

    def forward(self, x):
        if self.scale != 1:
            x = F.interpolate(x, scale_factor=1. / self.scale, mode="bilinear",
                              align_corners=False)
        x = self.conv0(x)
        x = self.convblock(x)
        x = self.conv1(x)
        flow = x
        if self.scale != 1:
            flow = F.interpolate(flow, scale_factor=self.scale, mode="bilinear",
                                 align_corners=False)
        return flow


class IFNet(nn.Module):
    def __init__(self):
        super(IFNet, self).__init__()
        self.block0 = IFBlock(6, scale=8, c=192)
        self.block1 = IFBlock(10, scale=4, c=128)
        self.block2 = IFBlock(10, scale=2, c=96)
        self.block3 = IFBlock(10, scale=1, c=48)

    def forward(self, x, UHD=False):
        if UHD:
            x = F.interpolate(x, scale_factor=0.5, mode="bilinear", align_corners=False)
        flow0 = self.block0(x)
        F1 = flow0
        F1_large = F.interpolate(F1, scale_factor=2.0, mode="bilinear", align_corners=False, recompute_scale_factor=False) * 2.0
        warped_img0 = warp(x[:, :3], F1_large[:, :2])
        warped_img1 = warp(x[:, 3:], F1_large[:, 2:4])
        flow1 = self.block1(torch.cat((warped_img0, warped_img1, F1_large), 1))
        F2 = (flow0 + flow1)
        F2_large = F.interpolate(F2, scale_factor=2.0, mode="bilinear", align_corners=False, recompute_scale_factor=False) * 2.0
        warped_img0 = warp(x[:, :3], F2_large[:, :2])
        warped_img1 = warp(x[:, 3:], F2_large[:, 2:4])
        flow2 = self.block2(torch.cat((warped_img0, warped_img1, F2_large), 1))
        F3 = (flow0 + flow1 + flow2)
        F3_large = F.interpolate(F3, scale_factor=2.0, mode="bilinear", align_corners=False, recompute_scale_factor=False) * 2.0
        warped_img0 = warp(x[:, :3], F3_large[:, :2])
        warped_img1 = warp(x[:, 3:], F3_large[:, 2:4])
        flow3 = self.block3(torch.cat((warped_img0, warped_img1, F3_large), 1))
        F4 = (flow0 + flow1 + flow2 + flow3)
        return F4, [F1, F2, F3, F4]

if __name__ == "__main__":
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    img0 = torch.zeros(3, 3, 256, 256).float().to(device)
    img1 = torch.tensor(np.random.normal(
        0, 1, (3, 3, 256, 256))).float().to(device)
    imgs = torch.cat((img0, img1), 1)
    flownet = IFNet()
    flow, _ = flownet(imgs)
    print(flow.shape)



# pylint: disable=all
import torch
import torch.nn as nn

backwarp_tenGrid = {}


def warp(tenInput: torch.Tensor, tenFlow: torch.Tensor):
    k = (str(tenFlow.device), str(tenFlow.size()))
    if k not in backwarp_tenGrid:
        tenHorizontal = torch.linspace(-1.0, 1.0, tenFlow.shape[3]).view(
            1, 1, 1, tenFlow.shape[3]).expand(tenFlow.shape[0], -1, tenFlow.shape[2], -1)
        tenVertical = torch.linspace(-1.0, 1.0, tenFlow.shape[2]).view(
            1, 1, tenFlow.shape[2], 1).expand(tenFlow.shape[0], -1, -1, tenFlow.shape[3])
        backwarp_tenGrid[k] = torch.cat(
            [tenHorizontal, tenVertical], 1).to(tenFlow.device)

    tenFlow = torch.cat([tenFlow[:, 0:1, :, :] / ((tenInput.shape[3] - 1.0) / 2.0),
                         tenFlow[:, 1:2, :, :] / ((tenInput.shape[2] - 1.0) / 2.0)], 1)

    g = (backwarp_tenGrid[k] + tenFlow).permute(0, 2, 3, 1)
    res = torch.nn.functional.grid_sample(input=tenInput, grid=torch.clamp(g, -1, 1),
                                          mode='bilinear', padding_mode='zeros', align_corners=True)
    return res



"""compute or load statistics for a set of Normed Representations of a MultiTaskDataset"""
from __future__ import annotations
from typing import Iterator
from pathlib import Path
import os
from torch.utils.data import DataLoader
import torch as tr
import numpy as np
from tqdm import tqdm

from vre.logger import vre_logger as logger
from vre.representations import NormedRepresentationMixin, Representation

TaskStatistics = tuple[tr.Tensor, tr.Tensor, tr.Tensor, tr.Tensor] # (min, max, mean, std)

def load_external_statistics(reader: "MultiTaskDataset",
                             statistics: dict[str, TaskStatistics | list]) -> dict[str, TaskStatistics]:
    """loads statistics from an external source provided to the constructor"""
    name_to_task: dict[str, Representation] = reader.name_to_task
    tasks_no_classif = [t for t in set(reader.task_names) if isinstance(name_to_task[t], NormedRepresentationMixin)]
    assert (diff := set(tasks_no_classif).difference(statistics)) == set(), f"Missing tasks: {diff}"
    res: dict[str, TaskStatistics] = {}
    for k, v in statistics.items():
        if k in reader.task_names:
            res[k] = tuple(tr.Tensor(x) for x in v)
            assert all(_stat.shape == (nd := (name_to_task[k].n_channels, )) for _stat in res[k]), (res[k], nd)
    logger.info(f"External statistics provided: { {k: tuple(v[0].shape) for k, v in res.items()} }")
    return res

def compute_statistics(reader: "MultiTaskDataset") -> dict[str, TaskStatistics]:
    """computes statistics for all tasks that are not classification"""
    name_to_task: dict[str, Representation] = reader.name_to_task
    cache_path: Path = reader.path / ".task_statistics.npz"
    res: dict[str, TaskStatistics] = {}
    if reader.cache_task_stats and cache_path.exists():
        res = np.load(cache_path, allow_pickle=True)["arr_0"].item()
        logger.info(f"Loaded task statistics: { {k: tuple(v[0].shape) for k, v in res.items()} } from {cache_path}")
    missing_tasks = [t for t in set(reader.task_names).difference(res)
                     if isinstance(name_to_task[t], NormedRepresentationMixin)]
    if len(missing_tasks) == 0:
        return res

    logger.info(f"Computing global task statistics (dataset len {len(reader)}) for {missing_tasks}")
    old_names, old_normalization = reader.task_names, reader.normalization
    reader.task_names, reader.normalization = missing_tasks, None # for reader[ix]
    _iter = iter(DataLoader(reader, batch_size=reader.batch_size_stats, num_workers=reader.num_workers_stats))
    res_missing = _c2(_iter, missing_tasks, reader.data_shape)
    reader.task_names, reader.normalization = old_names, old_normalization
    res = {**res, **res_missing}

    logger.info(f"Computed task statistics: { {k: tuple(v[0].shape) for k, v in res.items()} }")
    np.savez(cache_path, res)
    return res

def _c2(iterator: Iterator, tasks: list[str], data_shapes: dict[str, tuple[int, ...]]) -> dict[str, TaskStatistics]:
    # kinda based on: https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm
    # sadly this function cannot be at task-level because of speed issues, even if it'd be a bit more readable.
    def update(counts: tr.Tensor, counts_delta: float,  mean: tr.Tensor, m2: tr.Tensor,
                new_value: tr.Tensor) -> tuple[tr.Tensor, tr.Tensor, tr.Tensor]:
        new_count = counts + counts_delta
        batch_mean = new_value.nanmean(0)
        batch_var = ((new_value - batch_mean) ** 2).nansum(0)
        delta = batch_mean - mean
        new_count_no_zero = new_count + (new_count == 0) # add 1 (True) in case new_count is 0 to not divide by 0
        new_mean = mean + delta * counts_delta / new_count_no_zero
        new_m2 = m2 + batch_var + delta**2 * counts * counts_delta / new_count_no_zero
        assert not new_mean.isnan().any() and not new_m2.isnan().any(), (mean, new_mean, counts, counts_delta)
        return new_count, new_mean, new_m2

    ch = {task_name: data_shapes[task_name][-1] if len(data_shapes[task_name]) == 3 else 1 for task_name in tasks}
    counts = {task_name: tr.zeros(ch[task_name]).long() for task_name in tasks}
    mins = {task_name: tr.zeros(ch[task_name]).type(tr.float64) + 10**10 for task_name in tasks}
    maxs = {task_name: tr.zeros(ch[task_name]).type(tr.float64) - 10**10 for task_name in tasks}
    means_vec = {task_name: tr.zeros(ch[task_name]).type(tr.float64) for task_name in tasks}
    m2s_vec = {task_name: tr.zeros(ch[task_name]).type(tr.float64) for task_name in tasks}

    for (item, _) in tqdm(iterator, disable=os.getenv("STATS_PBAR", "0") == "0", desc="Computing stats"):
        assert all(task in item.keys() for task in tasks), f"Missing: {set(tasks).difference(item)}"
        for task in tasks:
            item_flat_ch = item[task].reshape(-1, ch[task])
            item_no_nan = item_flat_ch.nan_to_num(0).type(tr.float64)
            mins[task] = tr.minimum(mins[task], item_no_nan.min(0)[0])
            maxs[task] = tr.maximum(maxs[task], item_no_nan.max(0)[0])
            counts_delta = (item_flat_ch == item_flat_ch).long().sum(0) # pylint: disable=comparison-with-itself
            counts[task], means_vec[task], m2s_vec[task] = \
                update(counts[task], counts_delta, means_vec[task], m2s_vec[task], item_no_nan)

    res = {}
    for task in tasks:
        res[task] = (mins[task], maxs[task], means_vec[task], (m2s_vec[task] / counts[task]).sqrt())
        assert not any(x[0].isnan().any() for x in res[task]), (task, res[task])
    return res



"""Video Repentations Extractor module"""
from __future__ import annotations
from pathlib import Path
from datetime import datetime
import traceback
from tqdm import tqdm
import torch as tr
import numpy as np

from .representations import Representation, ComputeRepresentationMixin, LearnedRepresentationMixin
from .representations.io_representation_mixin import IORepresentationMixin
from .vre_runtime_args import VRERuntimeArgs
from .data_writer import DataWriter
from .data_storer import DataStorer
from .metadata import RunMetadata, RepresentationMetadata
from .vre_video import VREVideo
from .utils import now_fmt, make_batches, vre_topo_sort, ReprOut, DiskData, SummaryPrinter
from .logger import vre_logger as logger

def _add_data_writers_to_run_metadata(run_metadata, representations, output_dir, output_dir_exists_mode):
    for vrepr in representations:
        run_metadata.data_writers[vrepr.name] = DataWriter(output_dir=output_dir, representation=vrepr,
                                                           output_dir_exists_mode=output_dir_exists_mode).to_dict()

class VideoRepresentationsExtractor:
    """Video Representations Extractor class"""

    def __init__(self, video: VREVideo, representations: list[Representation]):
        """
        Parameters:
        - video The video we are performing VRE one
        - representations The list instantiated representations. Must be topo-sortable based on name and deps.
        """
        assert len(representations) > 0, "At least one representation must be provided"
        assert all(isinstance(x, Representation) for x in representations), [(x.name, type(x)) for x in representations]
        assert isinstance(video, VREVideo), (type(video), video)
        self.video = video
        self.representations: list[Representation] = vre_topo_sort(representations)
        self.repr_names = [r.name for r in representations]

    def set_compute_params(self, **kwargs) -> VideoRepresentationsExtractor:
        """Set the required params for all representations of ComputeRepresentationMixin type"""
        for r in [_r for _r in self.representations if isinstance(_r, ComputeRepresentationMixin)]:
            r.set_compute_params(**kwargs)
        return self

    def set_io_parameters(self, **kwargs) -> VideoRepresentationsExtractor:
        """Set the required params for all representations of IORepresentationMixin type"""
        for r in [_r for _r in self.representations if isinstance(_r, IORepresentationMixin)]:
            r.set_io_params(**kwargs)
        return self

    def to_graphviz(self, **kwargs) -> "Digraph":
        """Returns a graphviz object from this graph. Used for plotting the graph. Best for smaller graphs."""
        from graphviz import Digraph # pylint: disable=import-outside-toplevel
        g = Digraph()
        for k, v in kwargs.items():
            g.attr(**{k: v})
        g.attr(rankdir="LR")
        for node in self.representations:
            g.node(name=f"{node.name}", shape="oval")
        edges: list[tuple[str, str]] = [(r.name, dep.name) for r in self.representations for dep in r.dependencies]
        for l, r in edges:
            g.edge(r, l) # reverse?
        return g

    def run(self, output_dir: Path, frames: list[int] | None = None, output_dir_exists_mode: str = "raise",
            exception_mode: str = "stop_execution", n_threads_data_storer: int = 0,
            exported_representations: list[str] | None = None) -> RunMetadata:
        """
        The main loop of the VRE. This will run all the representations on the video and store results in the output_dir
        Parameters:
        - output_dir The directory where VRE will store the representations
        - frames The list of frames to run VRE for. If None, will export all the frames of the video
        - output_dir_exists_mode What to do if the output dir already exists. See DataWriter docstring. Default 'raise'.
        - exception_mode What to do when encountering an exception. It always writes the exception to 'exception.txt'
          - 'skip_representation' Will stop the run of the current representation and start the next one
          - 'stop_execution' (default) Will stop the execution of VRE
        - n_threads_data_storer The number of threads used by the DataStorer
        - exported_representations If set, only this subset of representations are exported, otherwise all of them
            based on these provided to the VRE constructor.
        Returns:
        - A RunMetadata object representing the run statistics for all representations of this run.
        """
        self._setup_logger(output_dir, now := now_fmt())
        try:
            self.to_graphviz().render(pth := f"{output_dir}/.logs/graph-{now}", format="png", cleanup=True)
            logger.info(f"Stored graphviz representation at: '{pth}.png'")
        except Exception as e:
            logger.error(e)

        runtime_args = VRERuntimeArgs(self.video, self.representations, frames, exception_mode, n_threads_data_storer)
        logger.info(runtime_args)
        run_metadata = RunMetadata(self.repr_names, runtime_args, output_dir / f".logs/run_metadata-{now}.json")
        summary_printer = SummaryPrinter(self.repr_names, runtime_args)
        _add_data_writers_to_run_metadata(run_metadata, self.representations, output_dir, output_dir_exists_mode)

        for vre_repr in self._get_output_representations(exported_representations): # checks are done inside get fn
            repr_metadata = self.do_one_representation(run_id=run_metadata.id, representation=vre_repr,
                                                       output_dir=output_dir,
                                                       output_dir_exists_mode=output_dir_exists_mode,
                                                       runtime_args=runtime_args,)
            if repr_metadata.run_had_exceptions and runtime_args.exception_mode == "stop_execution":
                raise RuntimeError(f"Representation '{vre_repr.name}' threw. "
                                   f"Check '{logger.get_file_handler().baseFilename}' for information")
            run_metadata.add_run_stats(repr_metadata)
            summary_printer.repr_metadatas[vre_repr.name] = repr_metadata
        print(summary_printer())
        return run_metadata

    def do_one_representation(self, run_id: str, representation: Representation, output_dir: Path,
                              output_dir_exists_mode: str, runtime_args: VRERuntimeArgs) -> RepresentationMetadata:
        """The loop of each representation. Returns a representation metadata with information about this repr's run"""
        data_writer = DataWriter(output_dir=output_dir, representation=representation,
                                 output_dir_exists_mode=output_dir_exists_mode)
        data_storer = DataStorer(data_writer=data_writer, n_threads=runtime_args.n_threads_data_storer)
        logger.info(f"Running {run_id=}:\n{representation}\n{data_storer}")
        rep: Representation | ComputeRepresentationMixin | IORepresentationMixin = representation
        repr_metadata = RepresentationMetadata(repr_name=representation.name, run_id=run_id,
                                               disk_location=data_writer.rep_out_dir / ".repr_metadata.json",
                                               frames=list(range(len(self.video))))
        rep.output_size = self.video.frame_shape[0:2] if rep.output_size == "video_shape" else rep.output_size

        relevant_frames = [f for f in runtime_args.frames if f not in repr_metadata.frames_computed()]
        logger.debug(f"Out of {len(runtime_args.frames)} total frames, "
                     f"{len(runtime_args.frames) - len(relevant_frames)} are precomputed and will be skipped.")
        batches = make_batches(relevant_frames, rep.batch_size)
        pbar = tqdm(total=len(relevant_frames), desc=f"[VRE] {rep.name} bs={rep.batch_size}")
        for batch in batches:
            now = datetime.now()
            if data_writer.all_batch_exists(batch):
                pbar.update(len(batch))
                continue

            try:
                tr.cuda.empty_cache() # might empty some unused memory, not 100% if needed.
                rep_data = self._compute_one_representation_batch(rep, batch=batch, output_dir=data_writer.output_dir)
                if not rep.is_classification and rep.name.find("fastsam") == -1: # TODO: MemoryData of M2F is binary...
                    assert rep_data.output.shape[-1] == rep.n_channels, (rep, rep_data.output, rep.n_channels)
                rep_data.output_images = rep.make_images(rep_data) if rep.export_image else None
                data_storer(rep_data)
            except Exception:
                self._log_error(f"\n[{rep.name} {rep.batch_size=} {batch=}] {traceback.format_exc()}\n")
                repr_metadata.add_time(None, batch)
                repr_metadata.run_had_exceptions = True
                break
            repr_metadata.add_time((datetime.now() - now).total_seconds(), batch)
            pbar.update(len(batch))

        # various cleanup stuff before ending with this representation
        data_storer.join_with_timeout(timeout=30)
        if isinstance(representation, LearnedRepresentationMixin) and representation.setup_called:
            representation.vre_free()
        for dep in representation.dependencies:
            if isinstance(dep, LearnedRepresentationMixin) and dep.setup_called:
                dep.vre_free()
        repr_metadata.store_on_disk()
        return repr_metadata

    # Helper private methods

    def _load_from_disk_if_possible(self, rep: Representation, video: VREVideo, ixs: list[int],
                                    output_dir: Path) -> ReprOut | None:
        """loads (batched) data from disk if possible."""
        assert isinstance(rep, IORepresentationMixin), rep
        assert isinstance(ixs, list) and all(isinstance(ix, int) for ix in ixs), (type(ixs), [type(ix) for ix in ixs])
        assert output_dir is not None and output_dir.exists(), output_dir
        # TODO: use data writer to create the paths as they can be non-npz for example
        npz_paths: list[Path] = [output_dir / rep.name / f"npz/{ix}.npz" for ix in ixs]
        extra_paths: list[Path] = [output_dir / rep.name / f"npz/{ix}_extra.npz" for ix in ixs]
        if any(not x.exists() for x in npz_paths): # partial batches are considered 'not existing' and overwritten
            return None
        extras_exist = [x.exists() for x in extra_paths]
        assert (ee := sum(extras_exist)) in (0, (ep := len(extra_paths))), f"Found {ee}. Expected either 0 or {ep}"
        disk_data: DiskData = np.array([rep.load_from_disk(x) for x in npz_paths])
        extra = [np.load(x, allow_pickle=True)["arr_0"].item() for x in extra_paths] if ee == ep else None
        logger.debug2(f"[{rep}] Slice: [{ixs[0]}:{ixs[-1]}]. All data found on disk and loaded")
        return ReprOut(frames=video[ixs], output=rep.disk_to_memory_fmt(disk_data), extra=extra, key=ixs)

    def _compute_one_representation_batch(self, rep: Representation, batch: list[int],
                                          output_dir: Path, depth: int = 0) -> ReprOut:
        assert isinstance(rep, ComputeRepresentationMixin), rep
        # setting depth to higher values allows to compute deps of deps in memory. We throw to not hide bugs.
        # useful for VRE hacking (i.e. see semantic_mapper.py)
        assert depth <= 1, f"{rep=} {depth=} {batch=} {output_dir=}"
        loaded_data = self._load_from_disk_if_possible(rep, self.video, batch, output_dir)
        if loaded_data is not None:
            return loaded_data
        dep_data = []
        for dep in rep.dependencies:
            dep_data.append(self._compute_one_representation_batch(dep, batch, output_dir, depth + 1))
        if isinstance(rep, LearnedRepresentationMixin) and rep.setup_called is False:
            rep.vre_setup() # instantiates the model, loads to cuda device etc.
        return rep.compute(self.video, ixs=batch, dep_data=dep_data)

    def _setup_logger(self, output_dir: Path, now_str: str):
        (logs_dir := output_dir / ".logs").mkdir(exist_ok=True, parents=True)
        logs_file = logs_dir / f"logs-{now_str}.txt"
        logger.add_file_handler(logs_file)
        logger.info(f"Logging run at: '{logs_file}")

    def _log_error(self, msg: str):
        logs_file = logger.get_file_handler().baseFilename
        open(logs_file, "a").write(f"{'=' * 80}\n{now_fmt()}\n{msg}\n{'=' * 80}")
        logger.debug(f"Error: {msg}")

    def _get_output_representations(self, exported_representations: list[str]) \
            -> list[IORepresentationMixin | Representation]:
        """given all the representations, keep those that actually export something. At least one is needed"""
        crs: list[IORepresentationMixin] = [_r for _r in self.representations if isinstance(_r, IORepresentationMixin)]
        assert len(crs) > 0, f"No I/O Representation found in {self.repr_names}"
        out_r: list[Representation] = [r for r in crs if r.export_binary or r.export_image]
        if exported_representations:
            logger.info(f"Explicit subset providewd: {exported_representations}. Exporting only these.")
            out_r = [r for r in out_r if r.name in exported_representations]

        assert len(out_r) > 0, f"No output format set for any I/O Representation: {', '.join([r.name for r in crs])}"
        for vre_repr in out_r:
            assert isinstance(vre_repr, Representation), vre_repr
            assert isinstance(vre_repr, IORepresentationMixin), vre_repr
            assert vre_repr.binary_format is not None or vre_repr.image_format is not None, vre_repr

        return out_r

    def __call__(self, *args, **kwargs) -> RunMetadata:
        return self.run(*args, **kwargs)

    def __str__(self) -> str:
        return f"""\n[VRE]
- Video: {self.video}
- Representations ({len(self.representations)}): [{", ".join(self.repr_names)}]"""

    def __repr__(self) -> str:
        return str(self)

    def __getitem__(self, key: str) -> Representation:
        assert key in self.repr_names, f"Representation '{key}' not in {self.repr_names}"
        return self.representations[self.repr_names.index(key)]



"""Weights utils. Stuff related to weights repository, downloading and loading weights"""
from pathlib import Path
from urllib.request import urlretrieve
import os

from .utils import get_project_root, DownloadProgressBar, is_git_lfs
from ..logger import vre_logger as logger

REPO_URL = "https://gitlab.com/video-representations-extractor/video-representations-extractor"
RESOURCES_URL = f"{REPO_URL}/-/raw/master/resources"
RESOURCES_DIR = Path(os.getenv("VRE_RESOURCES_DIR", get_project_root() / "resources"))

def fetch_resource(resource_name: str) -> Path:
    """fetches a resources from gitlab LFS if needed"""
    url = f"{RESOURCES_URL}/{resource_name}"
    (path := RESOURCES_DIR / resource_name).parent.mkdir(exist_ok=True, parents=True)
    if not Path(path).exists() or is_git_lfs(path):
        with DownloadProgressBar(unit="B", unit_scale=True, miniters=1, desc=resource_name) as t:
            try:
                urlretrieve(url, filename=path, reporthook=t.update_to)
            except Exception as e:
                logger.info(f"Failed to download '{url}' to '{path}'")
                raise e
    return path



"""topological sort module"""
from typing import TypeVar

T = TypeVar("T")

def topological_sort(dependency_graph: dict[T, list[T]]) -> list[T]:
    """
    Given a graph as a dict {Node : [Dependencies]}, returns a list [Node] ordered with a correct topological
        sort order. Applies Kahn's algorithm: https://ocw.cs.pub.ro/courses/pa/laboratoare/laborator-07
    """
    L, S = [], []

    # First step is to create a regular graph of {Node : [Children]}
    graph: dict[str, T] = {k: [] for k in dependency_graph.keys()}
    in_nodes_graph = {}
    for key in dependency_graph:
        for parent in dependency_graph[key]:
            assert parent in graph, f"Node '{parent}' is not in given graph: {graph.keys()}"
            graph[parent].append(key)
        # Transform the depGraph into a list of number of in-nodes
        in_nodes_graph[key] = len(dependency_graph[key])
        # Add nodes with no dependencies and start BFS from them
        if in_nodes_graph[key] == 0:
            S.append(key)

    while len(S) > 0:
        u = S.pop()
        L.append(u)

        for v in graph[u]:
            in_nodes_graph[v] -= 1
            if in_nodes_graph[v] == 0:
                S.insert(0, v)

    for key in in_nodes_graph.keys():
        assert in_nodes_graph[key] == 0, "Graph has cycles. Cannot do topological sort."
    return L

def vre_topo_sort(representations: list["Representation"]) -> list["Representation"]:
    """does a topological sort of a list of representations. They must have .name and .deps"""
    dep_graph = {r.name: [_r.name for _r in r.dependencies] for r in representations}
    tsr = topological_sort(dep_graph)

    # do a DFS to ensure that all the representations of the same are the same object!
    # guaranteed to have no cycles due to the topo sort above
    repr_to_id = {r.name: id(r) for r in representations}
    to_process = [*representations]
    while len(to_process) > 0:
        r = to_process.pop()
        if r.name in repr_to_id:
            if (id1 := id(r)) != (id2 := repr_to_id[r.name]):
                raise ValueError(f"Representation '{r}' has a different id: {id1} than previously provided: {id2}")
            to_process.extend(r.dependencies)
        else:
            repr_to_id[r.name] = id(r)

    name_to_repr = {r.name: r for r in representations}
    return [name_to_repr[r_name] for r_name in tsr]



"""Weights Repository module for Learnable Representations in the default repository"""
from pathlib import Path
from vre.utils.resources import fetch_resource

def fetch_weights(paths: list[str], depth: int = 0) -> list[Path]:
    """fetches weights for a representation from the repository (if needed) and returns the local path"""
    assert isinstance(paths, list), (paths, type(paths))
    assert depth <= 1, depth
    res = []
    for path in paths:
        if isinstance(path, list): # cases like marigold where 1 ckpt file is split across multiple ckpt files
            local_path = fetch_weights(path, depth + 1)[0].parent
        else:
            local_path = fetch_resource(f"weights/{path}")
        res.append(local_path)
    return res



"""summary printer at the end of a vre run"""
import numpy as np
from .utils import str_maxk

class SummaryPrinter:
    """summary printer at the end of a vre run"""
    def __init__(self, representations: list[str], runtime_args: "VRERuntimeArgs"):
        self.representations = representations
        self.repr_metadatas: dict[str, "RepresentationMetadata" | None] = {r: None for r in representations}
        self.runtime_args = runtime_args.to_dict()

    @property
    def run_stats(self) -> dict[str, dict[str, float]]:
        """The run stats of each representation that was registered to this run or older ones (only computed values)"""
        res = {}
        for r in self.repr_metadatas.values():
            if r is not None:
                _res = {k: v for k, v in r.run_stats.items() if v is not None}
                if len(_res) > 0:
                    res[r.repr_name] = _res
        return res

    def __call__(self) -> str:
        """returns a pretty formatted string of the metadata"""
        res = ""
        frames = self.runtime_args["frames"]
        chosen_frames = sorted(np.random.choice(frames, size=min(5, len(frames)), replace=False))
        res = f"{'Name':<20}" + "|" + "|".join([f"{str_maxk(str(k), 9):<9}" for k in chosen_frames]) + "|Total"
        for vrepr in self.representations:
            if vrepr not in self.run_stats:
                continue
            stats = np.array([x["duration"] or float("nan") for x in self.run_stats[vrepr].values()]).round(3)
            res += "\n" + f"{str_maxk(vrepr, 20):<20}"
            for chosen_frame in chosen_frames:
                res += "|" + f"{stats[frames.index(chosen_frame)]:<9}"
            res += "|" + f"{stats.sum().round(2)}"
        return res



"""SafeUAV Model implementation"""
# pylint: disable=all
from __future__ import annotations
import torch as tr
import torch.nn as nn
import torch.nn.functional as F

def conv(d_in: int, d_out: int, kernel_size: int | tuple[int, int], padding: int,
         stride: int, dilation: int, dropout: float) -> nn.Sequential:
    return nn.Sequential(
        nn.Conv2d(in_channels=d_in, out_channels=d_out, kernel_size=kernel_size, padding=padding,
                  stride=stride, dilation=dilation),
        nn.BatchNorm2d(num_features=d_out),
        nn.ReLU(inplace=True),
        nn.Dropout(dropout),
    )

def conv_t(d_in: int, d_out: int, kernel_size: int | tuple[int, int], padding: int,
            stride: int, dilation: int, dropout: float) -> nn.Sequential:
    return nn.Sequential(
        nn.ConvTranspose2d(in_channels=d_in, out_channels=d_out, kernel_size=kernel_size, padding=padding,
                           stride=stride, dilation=dilation),
        nn.BatchNorm2d(num_features=d_out),
        nn.ReLU(inplace=True),
        nn.Dropout(dropout),
    )

def cat_fn(a: tr.Tensor, b: tr.Tensor) -> tr.Tensor:
    diff_up, diff_left = b.shape[-2] - a.shape[-2], b.shape[-1] - a.shape[-1]
    a = F.pad(a, (0, diff_left, 0, diff_up))
    c = tr.cat([a, b], dim=1)
    return c

class EncoderMap2Map(nn.Module):
    def __init__(self, d_in: int, n_filters: int, dropout: float=0):
        super().__init__()
        self.d_in = d_in
        self.n_filters = n_filters
        self.dropout = dropout

        self.conv1 = conv(d_in=d_in, d_out=n_filters, kernel_size=3, padding=1,
                          stride=1, dilation=1, dropout=dropout)
        self.conv2 = conv(d_in=n_filters, d_out=n_filters, kernel_size=3, padding=1,
                          stride=1, dilation=1, dropout=dropout)
        self.conv3 = conv(d_in=n_filters, d_out=n_filters, kernel_size=3, padding=1,
                          stride=2, dilation=1, dropout=dropout)
        self.conv4 = conv(d_in=n_filters, d_out=n_filters * 2, kernel_size=3, padding=1,
                          stride=1, dilation=1, dropout=dropout)
        self.conv5 = conv(d_in=n_filters * 2, d_out=n_filters * 2, kernel_size=3, padding=1,
                          stride=1, dilation=1, dropout=dropout)
        self.conv6 = conv(d_in=n_filters * 2, d_out=n_filters * 2, kernel_size=3, padding=1,
                          stride=2, dilation=1, dropout=dropout)
        self.conv7 = conv(d_in=n_filters * 2, d_out=n_filters * 4, kernel_size=3, padding=1,
                          stride=1, dilation=1, dropout=dropout)
        self.conv8 = conv(d_in=n_filters * 4, d_out=n_filters * 4, kernel_size=3, padding=1,
                          stride=1, dilation=1, dropout=dropout)
        self.conv9 = conv(d_in=n_filters * 4, d_out=n_filters * 4, kernel_size=3, padding=1,
                          stride=2, dilation=1, dropout=dropout)

        self.dilate1 = conv(d_in=n_filters * 4, d_out=n_filters * 8, kernel_size=3, padding=1,
                            stride=1, dilation=1, dropout=dropout)
        self.dilate2 = conv(d_in=n_filters * 8, d_out=n_filters * 8, kernel_size=3, padding=2,
                            stride=1, dilation=2, dropout=dropout)
        self.dilate3 = conv(d_in=n_filters * 8, d_out=n_filters * 8, kernel_size=3, padding=4,
                            stride=1, dilation=4, dropout=dropout)
        self.dilate4 = conv(d_in=n_filters * 8, d_out=n_filters * 8, kernel_size=3, padding=8,
                            stride=1, dilation=8, dropout=dropout)
        self.dilate5 = conv(d_in=n_filters * 8, d_out=n_filters * 8, kernel_size=3, padding=16,
                            stride=1, dilation=16, dropout=dropout)
        self.dilate6 = conv(d_in=n_filters * 8, d_out=n_filters * 8, kernel_size=3, padding=32,
                            stride=1, dilation=32, dropout=dropout)

        self.conv10_t = conv_t(d_in=n_filters * 8, d_out=n_filters * 4, kernel_size=3, padding=1,
                               stride=2, dilation=1, dropout=dropout)
        self.conv11 = conv(d_in=n_filters * 4 * 2, d_out=n_filters * 4, kernel_size=3, padding=1,
                           stride=1, dilation=1, dropout=dropout)
        self.conv12_t = conv_t(d_in=n_filters * 4, d_out=n_filters * 2, kernel_size=3, padding=1,
                               stride=2, dilation=1, dropout=dropout)
        self.conv13 = conv(d_in=n_filters * 2 * 2, d_out=n_filters, kernel_size=3, padding=1,
                           stride=1, dilation=1, dropout=dropout)
        self.conv14_t = conv_t(d_in=n_filters, d_out=n_filters, kernel_size=3, padding=1,
                               stride=2, dilation=1, dropout=dropout)
        self.conv15 = conv(d_in=n_filters * 2, d_out=n_filters, kernel_size=3, padding=1,
                           stride=1, dilation=1, dropout=dropout)

    def forward(self, x: tr.Tensor, **kwargs) -> tr.Tensor:
        # x::d_inxHxW
        y_1 = self.conv1(x)
        y_2 = self.conv2(y_1)
        y_3 = self.conv3(y_2)
        y_4 = self.conv4(y_3)
        y_5 = self.conv5(y_4)
        y_6 = self.conv6(y_5)

        y_7 = self.conv7(y_6)
        y_8 = self.conv8(y_7)
        y_9 = self.conv9(y_8)

        y_dilate1 = self.dilate1(y_9)
        y_dilate2 = self.dilate2(y_dilate1)
        y_dilate3 = self.dilate3(y_dilate2)
        y_dilate4 = self.dilate4(y_dilate3)
        y_dilate5 = self.dilate5(y_dilate4)
        y_dilate6 = self.dilate6(y_dilate5)
        y_dilate_sum = y_dilate1 + y_dilate2 + y_dilate3 + y_dilate4 + y_dilate5 + y_dilate6

        y_10 = cat_fn(self.conv10_t(y_dilate_sum), y_7)
        y_11 = self.conv11(y_10)
        y_12 = cat_fn(self.conv12_t(y_11), y_4)
        y_13 = self.conv13(y_12)
        y_14 = cat_fn(self.conv14_t(y_13), y_1)
        y_15 = self.conv15(y_14)
        return y_15

class SafeUAV(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, num_filters: int):
        super().__init__()
        self.encoder = EncoderMap2Map(in_channels, num_filters)
        self.decoder = nn.Conv2d(in_channels=num_filters, out_channels=out_channels,
                                 kernel_size=1, padding=0, stride=1, dilation=1)

    def forward(self, x: tr.Tensor) -> tr.Tensor:
        y_encoder = self.encoder(x)
        y_decoder = self.decoder(y_encoder)
        return y_decoder



"""edges_representation.py -- module implementing an Edges Represenatation generic class"""
from overrides import overrides
import numpy as np
from vre.vre_video import VREVideo
from vre.utils import ReprOut
from vre.representations import (
    Representation, NpIORepresentation, ComputeRepresentationMixin, NormedRepresentationMixin)

class EdgesRepresentation(Representation, NpIORepresentation, ComputeRepresentationMixin, NormedRepresentationMixin):
    """EdgesRepresentation -- CV representation for 1-channeled edges/boundaries"""
    def __init__(self, name: str, **kwargs):
        Representation.__init__(self, name, **kwargs)
        ComputeRepresentationMixin.__init__(self)
        NpIORepresentation.__init__(self)
        NormedRepresentationMixin.__init__(self)

    @property
    @overrides
    def n_channels(self) -> int:
        return 1

    def make_images(self, data: ReprOut) -> np.ndarray:
        y = self.unnormalize(data.output) if self.normalization is not None else data.output
        return (np.repeat(y, 3, axis=-1) * 255).astype(np.uint8)

    @overrides
    def compute(self, video: VREVideo, ixs: list[int], dep_data: list[ReprOut] | None = None) -> ReprOut:
        raise NotImplementedError(f"[{self}] compute() must be overriden. We inherit it for output_dtype/size etc.")



"""FakeVideo module"""
from pathlib import Path
import shutil
from overrides import overrides
from tqdm import trange
import numpy as np
from .vre_video import VREVideo
from ..utils import image_write

class FakeVideo(VREVideo):
    """
    FakeVideo -- class used to test representations with a given numpy array.
    If 'frames' is provided, then these are the frames indexes, i.e. video[ix] returns self.data[self.frames_ix[ix]]
    This can be used to create a FakeVideo out of a directory with files: [1.png, 10.png, 100.png]
    """
    def __init__(self, data: np.ndarray, fps: float, frames: list[int] | None = None):
        self.data = data
        super().__init__()
        assert len(data) > 0, "No data provided"
        self._fps = fps
        self.frames = list(range(len(data))) if frames is None else frames
        self.path = f"FakeVideo {self.data.shape}"
        assert 0 < len(self.frames) <= 1_000_000 # max 1M frames to keep it tight
        assert len(self.frames) == len(self.data), (self.frames, self.data)
        assert all(isinstance(frame, int) for frame in self.frames), self.frames
        assert sorted(set(self.frames)) == self.frames, self.frames
        self.frames_ix = dict(zip(self.frames, range(len(self.frames)))) # {ix: frame}

    @property
    @overrides
    def shape(self) -> tuple[int, int, int, int]:
        return self.data.shape

    @property
    @overrides
    def fps(self) -> float:
        return self._fps

    @overrides
    def write(self, out_path: Path, start_frame: int = 0, end_frame: int | None = None):
        if (pth := Path(out_path)).exists():
            shutil.rmtree(out_path, ignore_errors=True)
        pth.mkdir(parents=True)
        for ix in trange((len(self) if end_frame is None else end_frame) - start_frame, desc=pth.name):
            out_file = pth / f"{(start_frame + ix) * 1 / self.fps}.png"
            image_write(self[start_frame + ix], out_file)

    @overrides
    def get_one_frame(self, ix: int) -> np.ndarray:
        return self.data[self.frames_ix[ix]]

    def __len__(self) -> int:
        return self.frames[-1] + 1

    def __repr__(self):
        return f"[FakeVideo] FPS: {self.fps}. Len: {len(self.data)}. Frame shape: {self.data.shape[1:]}. FPS: "



"""init file"""
from .utils import (get_project_root, parsed_str_type, is_dir_empty, get_closest_square, str_maxk, array_blend, now_fmt,
                    semantic_mapper, FixedSizeOrderedDict, abs_path, reorder_dict, make_batches, vre_load_weights, clip,
                    load_function_from_module, random_chars, mean)
from .topological_sort import topological_sort, vre_topo_sort
from .image import (image_resize, image_resize_batch, image_write, image_read,
                    image_add_title, collage_fn, image_add_border, image_blend)
from .resources import fetch_resource
from .colorizer import colorize_depth, colorize_semantic_segmentation, colorize_optical_flow
from .lovely import lo, monkey_patch
from .repr_memory_layout import ReprOut, MemoryData, DiskData
from .atomic_open import AtomicOpen
from .summary_printer import SummaryPrinter
from .yaml import vre_yaml_load



"""yaml utils -- a minimum subset of stuff from omegaconf"""
from io import IOBase, StringIO
from pathlib import Path
import os
from typing import Any
import yaml

def vre_yaml_load(path: Path | str | IOBase) -> dict[str, Any]:
    """reads a yaml file and resolves the env varaibles if needed"""
    fp = open(path, "r") if isinstance(path, (Path, str)) else path
    fp.seek(0)
    data: str = fp.read()
    skip = 0
    new_data = ""
    while (ix := data.find("${oc.env:", skip)) != -1:
        new_data += data[skip: ix]
        ix_r = data.find("}", ix+9)
        if ix_r == -1:
            raise ValueError(f"Unfinished oc.env: {data[ix:]}")
        key_and_default = data[ix+9: ix_r].split(",")
        assert len(key_and_default) <= 2, key_and_default
        has_default = False
        key, default = key_and_default[0], None
        if len(key_and_default) == 2:
            has_default = True
            key, default = key_and_default
        if not has_default and key not in os.environ:
            raise KeyError(f"Environment variable: '{key}' not set and no default provided.")
        if "{" in key:
            raise ValueError("'{' cannot be in the read key: '" + key + "'")
        read_key = os.getenv(key, default)
        new_data += read_key
        skip = ix_r + 1
    new_data += data[skip:]
    cfg = yaml.safe_load(StringIO(new_data))
    return cfg



"""lovely_numpy/tensors reimplementation"""
import torch as tr
import numpy as np

def monkey_patch(cls=tr.Tensor):
    "Monkey-patch lovely features into `cls`"
    setattr(cls, "v", cls.__repr__)
    cls.__repr__ = cls.__str__ = lo

def lo(x: np.ndarray | tr.Tensor | None) -> str:
    """reimplementation of lovely_numpy's lo() without any extra stuff that spams the terminal. Only for numericals!"""
    def _dt(x: tr.dtype | np.dtype) -> str:
        y = str(x).removeprefix("torch.") if str(x).startswith("torch.") else str(x)
        return y if y == "bool" else (f"{y[0]}{y[-1]}" if y[-1] == "8" else f"{y[0]}{y[-2:]}")
    assert isinstance(x, (type(None), np.ndarray, tr.Tensor)), type(x)
    if x is None:
        return x
    _r = lambda x: round(x.item(), 2) # pylint: disable=unnecessary-lambda-assignment
    arr_type = "arr" if isinstance(x, np.ndarray) else "tr"
    μ, σ = x.float().mean() if arr_type == "tr" else x.mean(), x.float().std() if arr_type == "tr" else x.std() # thx tr
    return f"{arr_type}{[*x.shape]} {_dt(x.dtype)} x∈[{_r(x.min())}, {_r(x.max())}], μ={_r(μ)}, σ={_r(σ)}"



# pylint: disable=all
from .model import FastSAM
import numpy as np
from PIL import Image
from typing import Optional


class FastSAMDecoder:
    def __init__(
        self,
        model: FastSAM,
        device: str='cpu',
        conf: float=0.4,
        iou: float=0.9,
        imgsz: int=1024,
        retina_masks: bool=True,
        ):
        self.model = model
        self.device = device
        self.retina_masks = retina_masks
        self.imgsz = imgsz
        self.conf = conf
        self.iou = iou
        self.image = None
        self.image_embedding = None

    def run_encoder(self, image):
        if isinstance(image,str):
            image =  np.array(Image.open(image))
        self.image = image
        image_embedding = self.model(
            self.image,
            device=self.device,
            retina_masks=self.retina_masks,
            imgsz=self.imgsz,
            conf=self.conf,
            iou=self.iou
            )
        return image_embedding[0].numpy()

    def run_decoder(
            self,
            image_embedding,
            point_prompt: Optional[np.ndarray]=None,
            point_label: Optional[np.ndarray]=None,
            box_prompt: Optional[np.ndarray]=None,
            text_prompt: Optional[str]=None,
            )->np.ndarray:
        self.image_embedding = image_embedding
        if point_prompt is not None:
            ann = self.point_prompt(points=point_prompt, pointlabel=point_label)
            return ann
        elif box_prompt is not None:
            ann = self.box_prompt(bbox=box_prompt)
            return ann
        elif text_prompt is not None:
            ann = self.text_prompt(text=text_prompt)
            return ann
        else:
            return None

    def box_prompt(self, bbox):
        assert (bbox[2] != 0 and bbox[3] != 0)
        masks = self.image_embedding.masks.data
        target_height = self.image.shape[0]
        target_width = self.image.shape[1]
        h = masks.shape[1]
        w = masks.shape[2]
        if h != target_height or w != target_width:
            bbox = [
                int(bbox[0] * w / target_width),
                int(bbox[1] * h / target_height),
                int(bbox[2] * w / target_width),
                int(bbox[3] * h / target_height), ]
        bbox[0] = round(bbox[0]) if round(bbox[0]) > 0 else 0
        bbox[1] = round(bbox[1]) if round(bbox[1]) > 0 else 0
        bbox[2] = round(bbox[2]) if round(bbox[2]) < w else w
        bbox[3] = round(bbox[3]) if round(bbox[3]) < h else h

        # IoUs = torch.zeros(len(masks), dtype=torch.float32)
        bbox_area = (bbox[3] - bbox[1]) * (bbox[2] - bbox[0])

        masks_area = np.sum(masks[:, bbox[1]:bbox[3], bbox[0]:bbox[2]], axis=(1, 2))
        orig_masks_area = np.sum(masks, axis=(1, 2))

        union = bbox_area + orig_masks_area - masks_area
        IoUs = masks_area / union
        max_iou_index = np.argmax(IoUs)

        return np.array([masks[max_iou_index].cpu().numpy()])

    def point_prompt(self, points, pointlabel):  # numpy

        masks = self._format_results(self.image_embedding[0], 0)
        target_height = self.image.shape[0]
        target_width = self.image.shape[1]
        h = masks[0]['segmentation'].shape[0]
        w = masks[0]['segmentation'].shape[1]
        if h != target_height or w != target_width:
            points = [[int(point[0] * w / target_width), int(point[1] * h / target_height)] for point in points]
        onemask = np.zeros((h, w))
        masks = sorted(masks, key=lambda x: x['area'], reverse=True)
        for i, annotation in enumerate(masks):
            if type(annotation) == dict:
                mask = annotation['segmentation']
            else:
                mask = annotation
            for i, point in enumerate(points):
                if mask[point[1], point[0]] == 1 and pointlabel[i] == 1:
                    onemask[mask] = 1
                if mask[point[1], point[0]] == 1 and pointlabel[i] == 0:
                    onemask[mask] = 0
        onemask = onemask >= 1
        return np.array([onemask])

    def _format_results(self, result, filter=0):
        annotations = []
        n = len(result.masks.data)
        for i in range(n):
            annotation = {}
            mask = result.masks.data[i] == 1.0

            if np.sum(mask) < filter:
                continue
            annotation['id'] = i
            annotation['segmentation'] = mask
            annotation['bbox'] = result.boxes.data[i]
            annotation['score'] = result.boxes.conf[i]
            annotation['area'] = annotation['segmentation'].sum()
            annotations.append(annotation)
        return annotations



"""colorize semantic segmentation module -- based on the original M2F implementation but without matplotlib"""
import pycocotools.mask as mask_util
from PIL import Image, ImageDraw
import numpy as np

from ..lovely import lo
from ..pil_utils import _get_default_font, _pil_image_draw_textsize
from ..image import image_blend
from ..cv2_utils import cv2_findContours, cv2_RETR_CCOMP, cv2_CHAIN_APPROX_NONE, cv2_connectedComponentsWithStats

_AREA_THRESHOLD = 10
_LINE_WIDTH = 1.2
_LARGE_MASK_AREA_THRESH = 120_000
_WHITE = (255, 255, 255)
COLOR = tuple[int, int, int]

def colorize_semantic_segmentation(semantic_map: np.ndarray, classes: list[str], color_map: list[tuple[int, int, int]],
                                   rgb: np.ndarray | None = None, alpha: float = 0.8, size_px: int | None = None):
    """Colorize asemantic segmentation maps. Must be argmaxed (H, W). Can paint over the original RGB frame or not."""
    assert np.issubdtype(semantic_map.dtype, np.integer), semantic_map.dtype
    assert (max_class := semantic_map.max()) <= len(color_map), (max_class, len(color_map))
    assert len(shp := semantic_map.shape) == 3, shp
    assert rgb is None or (rgb.shape[0:-1] == shp), (rgb.shape, shp)
    alpha = alpha if rgb is not None else 1
    rgb = rgb if rgb is not None else np.zeros((*semantic_map.shape, 3), dtype=np.uint8)
    return np.array([_colorize_sem_seg(sema=_s, rgb=_r, classes=classes, color_map=color_map,
                                       alpha=alpha, size_px=size_px)
                     for _r, _s in zip(rgb, semantic_map)])

class _GenericMask:
    """
    Attribute:
        polygons (list[ndarray]): list[ndarray]: polygons for this mask.
            Each ndarray has format [x, y, x, y, ...]
        mask (ndarray): a binary mask
    """

    def __init__(self, mask, height, width):
        self._mask = self._polygons = self._has_holes = None
        self.height = height
        self.width = width

        assert isinstance(mask, np.ndarray), type(mask)
        assert mask.shape[1] != 2, mask.shape
        assert mask.shape == (height, width,), f"mask shape: {mask.shape}, target dims: {height}, {width}"
        self.mask = mask.astype(np.uint8)
        self.polygons, self.has_holes = self._mask_to_polygons(self.mask)

    def _mask_to_polygons(self, mask) -> tuple[list[tuple[int, int]], bool]:
        # cv2_RETR_CCOMP flag retrieves all the contours and arranges them to a 2-level
        # hierarchy. External contours (boundary) of the object are placed in hierarchy-1.
        # Internal contours (holes) are placed in hierarchy-2.
        # cv2_CHAIN_APPROX_NONE flag gets vertices of polygons from contours.
        mask = np.ascontiguousarray(mask)  # some versions of cv2 does not support incontiguous arr
        res = cv2_findContours(mask.astype("uint8"), cv2_RETR_CCOMP, cv2_CHAIN_APPROX_NONE)
        hierarchy = res[-1]
        if hierarchy is None:  # empty mask
            return [], False
        has_holes = (hierarchy.reshape(-1, 4)[:, 3] >= 0).sum() > 0
        res = res[-2]
        res = [x.flatten() for x in res]
        # These coordinates from OpenCV are integers in range [0, W-1 or H-1].
        # We add 0.5 to turn them into real-value coordinate space. A better solution
        # would be to first +0.5 and then dilate the returned polygon by 0.5.
        res = [x + 0.5 for x in res if len(x) >= 6]
        return res, has_holes

def _add_text_with_background(image_array, text, position, font_size, color):
    # Convert numpy array to PIL Image
    image = Image.fromarray(image_array)
    draw = ImageDraw.Draw(image)
    font = _get_default_font(font_size)

    text_width, text_height = _pil_image_draw_textsize(draw, text, font=font) # Calculate text size
    x, y = position # Define rectangle coordinates
    x = max(5, min(x, image_array.shape[1] - text_width - 5))
    y = max(5, min(y, image_array.shape[0] - text_height - 5))
    rectangle_coords = [x - 1, y + 2, x + text_width + 1, y + text_height]

    draw.rectangle(rectangle_coords, fill=(0, 0, 0)) # Draw black rectangle
    draw.text((x, y), text, fill=color, font=font) # add the text
    return np.array(image)

def _clip(a, left, right):
    assert left < right, (left, right)
    return max(left, min(right, a))

def _font_size_from_shape(shp) -> int:
    return _clip(min(*shp) // 15, 15, 30) # heuristic of the year

def _draw_text_in_mask(res: np.ndarray, binary_mask: np.ndarray, text: str, color: tuple[int, int, int],
                       size_px: int | None = None) -> np.ndarray:
    """
    Find proper places to draw text given a binary mask.
    """
    _num_cc, cc_labels, stats, _ = cv2_connectedComponentsWithStats(binary_mask, 8)
    if stats[1:, -1].size == 0:
        return res
    largest_component_id = np.argmax(stats[1:, -1]) + 1

    # draw text on the largest component, as well as other very large components.
    font_size_px =_font_size_from_shape(binary_mask.shape) if size_px is None else size_px
    for cid in range(1, _num_cc):
        if cid == largest_component_id or stats[cid, -1] > _LARGE_MASK_AREA_THRESH:
            center = np.median((cc_labels == cid).nonzero(), axis=1)[::-1] # position based on median
            res = _add_text_with_background(res, text, center, font_size_px, color)
    return res

def _draw_polygon(vertices: list[tuple[int, int]], shape: tuple[int, int], linewidth: int=1):
    """
    Draw a filled polygon with specified edge linewidth into a binary NumPy array, avoiding loops.

    :param vertices: A list or array of (x, y) tuples representing the polygon vertices.
    :param shape: A tuple (height, width) specifying the size of the output array.
    :param linewidth: The thickness of the polygon edges.
    :return: A binary NumPy array with the polygon drawn with filled interior and specified edge width.
    """
    vertices = np.array(vertices, dtype=np.float32)
    # expand both to fllor and ceil because we get values between 2 pixels. This is the easiest way to do a polygon
    # without "rasterizing" between the pixels. We also know that the vertices are of correct shape, so no need to
    # resize them.
    vertices = np.concatenate([np.floor(vertices), np.ceil(vertices)]).astype(float)
    radius = linewidth / 4.0 # / 4, not / 2 because of the +/-1 above

    # Calculate offsets for edge thickening
    edges = np.concatenate([vertices, vertices[:1]])  # Close the polygon
    edge_vectors = edges[1:] - edges[:-1]
    edge_lengths = np.linalg.norm(edge_vectors, axis=1, keepdims=True)
    normals = np.stack([-edge_vectors[:, 1], edge_vectors[:, 0]], axis=1)  # Perpendicular vector
    normals /= np.maximum(edge_lengths, 1e-6)  # Normalize to unit vectors

    # Expand polygon by the radius in both directions along normals
    expanded_vertices = []
    for direction in [-1, 1]:
        offset = direction * radius * normals
        expanded_vertices.append(edges[:-1] + offset)

    expanded_vertices = np.concatenate(expanded_vertices, axis=0)
    expanded_vertices[:, 0] = expanded_vertices[:, 0].clip(0, shape[1] - 1)
    expanded_vertices[:, 1] = expanded_vertices[:, 1].clip(0, shape[0] - 1)
    expanded_vertices = np.round(expanded_vertices).astype(int)

    # Rasterize expanded polygon
    binary_array = np.zeros(shape, dtype=bool)
    binary_array[expanded_vertices[:, 1], expanded_vertices[:, 0]] = 1
    return binary_array

def _colorize_sem_seg(sema: np.ndarray, rgb: np.ndarray | None, classes: list[str],
                      color_map: list[tuple[int, int, int]], alpha: float=0.8,
                      size_px: int | None = None) -> np.ndarray:
    """colorize semantic segmentation -- based on Mask2Former's approach but without MPL"""
    classes = list(map(str, classes)) if all(isinstance(x, int) for x in classes) else classes
    assert all(isinstance(x, str) for x in classes), classes
    sema_rgb = np.zeros((*sema.shape, 3), dtype=np.uint8)
    labels, areas = np.unique(sema, return_counts=True)
    sorted_idxs = np.argsort(-areas).tolist()
    sorted_labels = labels[sorted_idxs]
    res = sema_rgb
    masks: list[_GenericMask] = []
    alpha_all_masks = np.full((*rgb.shape[0:2], 1), 0).astype(np.float32) # an array jost for alphas per mask
    polys = np.zeros(rgb.shape[0:2], dtype=bool) # polygons binary mask that is added at the end
    texts_data: list[tuple[np.ndarray, str, COLOR]] = [] # a list of all texts based on the class labels (if shown)
    for ix in sorted_idxs:
        masks.append(_GenericMask(sema == labels[ix], *sema.shape[0:2]))

    for ix, mask in enumerate(masks):
        res[mask.mask.astype(bool)] = color_map[sorted_labels[ix]] # draw mask
        if mask.has_holes:
            texts_data.append((mask.mask, classes[sorted_labels[ix]], _WHITE)) # texts are in fg
        else:
            for segment in mask.polygons:
                area = mask_util.area(mask_util.frPyObjects([segment], *sema.shape[0:2]))
                if area < _AREA_THRESHOLD:
                    continue
                texts_data.append((mask.mask, classes[sorted_labels[ix]], _WHITE)) # texts are in fg
                polys = polys | _draw_polygon(segment.reshape(-1, 2), sema.shape[0:2], linewidth=_LINE_WIDTH)

        if rgb is not None:
            alpha_all_masks[mask.mask.astype(bool)] = 1 - alpha

    # apply all the extra transformations: rgb+alpha overlap -> white polygons -> texts on top of the masks
    if rgb is not None:
        assert rgb.dtype == np.uint8, lo(rgb)
        res = image_blend(res, rgb, alpha_all_masks)
    res[polys] = _WHITE
    for text_data in texts_data: # TODO: if texts overlap, show just the one with highest area maybe
        res = _draw_text_in_mask(res, binary_mask=text_data[0], text=text_data[1], color=text_data[2], size_px=size_px)
    return res



"""Metadata module -- The metadata of a particular VRE run"""
from __future__ import annotations
import json
from io import FileIO
from typing import Any
from pathlib import Path

from .utils import AtomicOpen, random_chars, mean
from .vre_runtime_args import VRERuntimeArgs

class RunMetadata:
    """Metadata of a run for multiple representations. Backed on the disk by a JSON file"""
    def __init__(self, representations: list[str], runtime_args: VRERuntimeArgs, disk_location: Path):
        assert len(representations) > 0 and all(isinstance(x, str) for x in representations), representations
        self.representations = representations
        self.disk_location = disk_location
        self.runtime_args = runtime_args.to_dict()
        self.id = random_chars(n=10)
        self.data_writers = {}
        self.run_stats = {}
        self.store_on_disk()

    @property
    def metadata(self) -> dict[str, Any]:
        """the run metadata as a json"""
        return {
            "id": self.id,
            "runtime_args": self.runtime_args,
            "data_writers": self.data_writers,
            "run_stats": self.run_stats,
        }

    def add_run_stats(self, repr_metadata: RepresentationMetadata):
        """adds statistics of a single representation after it finished running"""
        assert (name := repr_metadata.repr_name) not in self.run_stats, f"{name} in {self.run_stats.keys()}"
        frames_computed = repr_metadata.frames_computed(this_run_only=True)
        frames_failed = repr_metadata.frames_failed(this_run_only=True)

        avg_duration = round(mean([repr_metadata.run_stats[ix]["duration"] for ix in frames_computed]), 2)
        self.run_stats[name] = {"n_computed": len(frames_computed), "n_failed": len(frames_failed),
                                "average_duration": avg_duration}
        self.store_on_disk()

    def store_on_disk(self):
        """stores (overwrites if needed) the metadata on disk"""
        with open(self.disk_location, "w") as fp:
            json.dump(self.metadata, fp, indent=4)

    def __repr__(self):
        return (f"[Metadata] Id: {self.id}, Num representations: {len(self.representations)}. "
                f"Disk location: '{self.disk_location}'")

class RepresentationMetadata:
    """
    A class that defines the metadata and events that happened during a single representation's run.
    It is backed by a single sqlite file living in the representation's directory under .run_metadata.json.
    Note: that this file may be updated from multiple processes running at the same time. For this reason, we have an
    extra layer protection based on json modification time via `os.path.getmtime` and we merge before storing to disk.
    """
    def __init__(self, repr_name: str, run_id: str, disk_location: Path, frames: list[int]):
        assert all(isinstance(x, int) for x in frames), frames
        self.run_id = run_id
        self.repr_name = repr_name
        self.disk_location = disk_location
        self.run_had_exceptions = False
        self.run_stats: dict[int, float | None] = {f: None for f in frames}
        self.store_on_disk()

    def frames_computed(self, this_run_only: bool=False) -> list[int]:
        """returns the list of comptued frames so far for this representation"""
        return [ix for ix, v in self.run_stats.items() if v is not None and v["duration"] is not None
                and (v["run_id"] == self.run_id if this_run_only else True)]

    def frames_failed(self, this_run_only: bool=False) -> list[int]:
        """returns the list of failed frames so far for this representation"""
        return [ix for ix, v in self.run_stats.items() if v is not None and v["duration"] is None
                and (v["run_id"] == self.run_id if this_run_only else True)]

    def add_time(self, duration: float | None, frames: list[int]):
        """adds a (batched) time to the representation's run_stats"""
        assert (batch_size := len(frames)) > 0, batch_size
        data = [duration / batch_size] * batch_size if duration is not None else [None] * batch_size
        for ix, frame_duration in zip(frames, data):
            if self.run_stats[ix] is not None and self.run_stats[ix]["duration"] is not None:
                raise ValueError(f"Adding time to existing metadata {self}. Frame={ix}. Previous: {self.run_stats[ix]}")
            self.run_stats[ix] = {"run_id": self.run_id, "duration": frame_duration}
        self.store_on_disk()

    def store_on_disk(self):
        """
        Stores (overwrites) the metadata on disk. Note: this does allows for multi processing on the same repr!
        Note: that disk_writer_meta will be overwritten by the last writer.
        """
        file_exists = self.disk_location.exists()
        with AtomicOpen(self.disk_location, "a+") as fp:
            if file_exists: # if it exists, we merge the disk data with existing data before overwriting
                self.run_stats = self._load_and_merge_with_metadata_from_disk(fp)

            fp.seek(0)
            fp.truncate()
            json_data = {
                "name": self.repr_name,
                "run_stats": self.run_stats,
            }
            json.dump(json_data, fp, indent=4)

    def _load_and_merge_with_metadata_from_disk(self, fp: FileIO) -> dict[str, Any]:
        """given an (mutexed) fp, read the metadata from the disk and return the run stats. Fail if any errors"""
        fp.seek(0)
        loaded_json_data = json.loads(fp.read())
        loaded_run_stats = {int(k): v for k, v in loaded_json_data["run_stats"].items()} # convert from json keys to int
        assert (a := loaded_run_stats.keys()) == (b := self.run_stats.keys()), f"\n- {a}\n- {b}"
        assert (a := loaded_json_data["name"]) == (b := self.repr_name), f"\n- {a}\n- {b}"
        merged_run_stats = {}
        for k, v in self.run_stats.items():
            if (loaded_v := loaded_run_stats[k]) is not None and loaded_v["duration"] is not None:
                assert v == loaded_v or v is None, (k, v, loaded_v)
                merged_run_stats[k] = loaded_v
            else:
                merged_run_stats[k] = v
        return merged_run_stats

    def __repr__(self):
        return (f"[ReprMetadata] Representation: {self.repr_name}. Frames: {len(self.run_stats)} "
                f"(computed: {len(self.frames_computed())}, failed: {len(self.frames_failed())}) "
                f"Disk location: '{self.disk_location}'")



# pylint: disable=all
import numpy as np
import torch
from PIL import Image
from vre.logger import vre_logger as logger
import math
from types import SimpleNamespace
import re
import yaml
from pathlib import Path

class IterableSimpleNamespace(SimpleNamespace):
    """
    Ultralytics IterableSimpleNamespace is an extension class of SimpleNamespace that adds iterable functionality and
    enables usage with dict() and for loops.
    """

    def __iter__(self):
        """Return an iterator of key-value pairs from the namespace's attributes."""
        return iter(vars(self).items())

    def __str__(self):
        """Return a human-readable string representation of the object."""
        return '\n'.join(f'{k}={v}' for k, v in vars(self).items())

    def get(self, key, default=None):
        """Return the value of the specified key if it exists; otherwise, return the default value."""
        return getattr(self, key, default)

DEFAULT_CFG_DICT = {'task': 'detect', 'mode': 'train', 'model': None, 'data': None, 'epochs': 100, 'patience': 50, 'batch': 16, 'imgsz': 640, 'save': True, 'save_period': -1, 'cache': False, 'device': None, 'workers': 8, 'project': None, 'name': None, 'exist_ok': False, 'pretrained': True, 'optimizer': 'auto', 'verbose': True, 'seed': 0, 'deterministic': True, 'single_cls': False, 'rect': False, 'cos_lr': False, 'close_mosaic': 0, 'resume': False, 'amp': True, 'fraction': 1.0, 'profile': False, 'overlap_mask': True, 'mask_ratio': 4, 'dropout': 0.0, 'val': True, 'split': 'val', 'save_json': False, 'save_hybrid': False, 'conf': None, 'iou': 0.7, 'max_det': 300, 'half': False, 'dnn': False, 'plots': True, 'source': None, 'show': False, 'save_txt': False, 'save_conf': False, 'save_crop': False, 'show_labels': True, 'show_conf': True, 'vid_stride': 1, 'line_width': None, 'visualize': False, 'augment': False, 'agnostic_nms': False, 'classes': None, 'retina_masks': False, 'boxes': True, 'format': 'torchscript', 'keras': False, 'optimize': False, 'int8': False, 'dynamic': False, 'simplify': False, 'opset': None, 'workspace': 4, 'nms': False, 'lr0': 0.01, 'lrf': 0.01, 'momentum': 0.937, 'weight_decay': 0.0005, 'warmup_epochs': 3.0, 'warmup_momentum': 0.8, 'warmup_bias_lr': 0.1, 'box': 7.5, 'cls': 0.5, 'dfl': 1.5, 'pose': 12.0, 'kobj': 1.0, 'label_smoothing': 0.0, 'nbs': 64, 'hsv_h': 0.015, 'hsv_s': 0.7, 'hsv_v': 0.4, 'degrees': 0.0, 'translate': 0.1, 'scale': 0.5, 'shear': 0.0, 'perspective': 0.0, 'flipud': 0.0, 'fliplr': 0.5, 'mosaic': 1.0, 'mixup': 0.0, 'copy_paste': 0.0, 'cfg': None, 'v5loader': False, 'tracker': 'botsort.yaml'}
DEFAULT_CFG = IterableSimpleNamespace(**DEFAULT_CFG_DICT)
DEFAULT_CFG_KEYS = list(DEFAULT_CFG_DICT.keys())


# Define keys for arg type checks
CFG_FLOAT_KEYS = 'warmup_epochs', 'box', 'cls', 'dfl', 'degrees', 'shear'
CFG_FRACTION_KEYS = ('dropout', 'iou', 'lr0', 'lrf', 'momentum', 'weight_decay', 'warmup_momentum', 'warmup_bias_lr',
                     'label_smoothing', 'hsv_h', 'hsv_s', 'hsv_v', 'translate', 'scale', 'perspective', 'flipud',
                     'fliplr', 'mosaic', 'mixup', 'copy_paste', 'conf', 'iou', 'fraction')  # fraction floats 0.0 - 1.0
CFG_INT_KEYS = ('epochs', 'patience', 'batch', 'workers', 'seed', 'close_mosaic', 'mask_ratio', 'max_det', 'vid_stride',
                'line_width', 'workspace', 'nbs', 'save_period')
CFG_BOOL_KEYS = ('save', 'exist_ok', 'verbose', 'deterministic', 'single_cls', 'rect', 'cos_lr', 'overlap_mask', 'val',
                 'save_json', 'save_hybrid', 'half', 'dnn', 'plots', 'show', 'save_txt', 'save_conf', 'save_crop',
                 'show_labels', 'show_conf', 'visualize', 'augment', 'agnostic_nms', 'retina_masks', 'boxes', 'keras',
                 'optimize', 'int8', 'dynamic', 'simplify', 'nms', 'v5loader', 'profile')

def yaml_load(file='data.yaml', append_filename=False):
    """
    Load YAML data from a file.

    Args:
        file (str, optional): File name. Default is 'data.yaml'.
        append_filename (bool): Add the YAML filename to the YAML dictionary. Default is False.

    Returns:
        (dict): YAML data and file name.
    """
    with open(file, errors='ignore', encoding='utf-8') as f:
        s = f.read()  # string

        # Remove special characters
        if not s.isprintable():
            s = re.sub(r'[^\x09\x0A\x0D\x20-\x7E\x85\xA0-\uD7FF\uE000-\uFFFD\U00010000-\U0010ffff]+', '', s)

        # Add YAML filename to dict and return
        return {**yaml.safe_load(s), 'yaml_file': str(file)} if append_filename else yaml.safe_load(s)

def cfg2dict(cfg):
    """
    Convert a configuration object to a dictionary, whether it is a file path, a string, or a SimpleNamespace object.

    Args:
        cfg (str | Path | SimpleNamespace): Configuration object to be converted to a dictionary.

    Returns:
        cfg (dict): Configuration object in dictionary format.
    """
    if isinstance(cfg, (str, Path)):
        cfg = yaml_load(cfg)  # load dict
    elif isinstance(cfg, SimpleNamespace):
        cfg = vars(cfg)  # convert to dict
    return cfg

def get_cfg(cfg=DEFAULT_CFG_DICT, overrides: dict = None):
    """
    Load and merge configuration data from a file or dictionary.

    Args:
        cfg (str | Path | Dict | SimpleNamespace): Configuration data.
        overrides (str | Dict | optional): Overrides in the form of a file name or a dictionary. Default is None.

    Returns:
        (SimpleNamespace): Training arguments namespace.
    """
    cfg = cfg2dict(cfg)

    # Merge overrides
    if overrides:
        overrides = cfg2dict(overrides)
        cfg = {**cfg, **overrides}  # merge cfg and overrides dicts (prefer overrides)

    # Special handling for numeric project/name
    for k in 'project', 'name':
        if k in cfg and isinstance(cfg[k], (int, float)):
            cfg[k] = str(cfg[k])
    if cfg.get('name') == 'model':  # assign model to 'name' arg
        cfg['name'] = cfg.get('model', '').split('.')[0]
        logger.warning(f"WARNING ⚠️ 'name=model' automatically updated to 'name={cfg['name']}'.")

    # Type and Value checks
    for k, v in cfg.items():
        if v is not None:  # None values may be from optional args
            if k in CFG_FLOAT_KEYS and not isinstance(v, (int, float)):
                raise TypeError(f"'{k}={v}' is of invalid type {type(v).__name__}. "
                                f"Valid '{k}' types are int (i.e. '{k}=0') or float (i.e. '{k}=0.5')")
            elif k in CFG_FRACTION_KEYS:
                if not isinstance(v, (int, float)):
                    raise TypeError(f"'{k}={v}' is of invalid type {type(v).__name__}. "
                                    f"Valid '{k}' types are int (i.e. '{k}=0') or float (i.e. '{k}=0.5')")
                if not (0.0 <= v <= 1.0):
                    raise ValueError(f"'{k}={v}' is an invalid value. "
                                     f"Valid '{k}' values are between 0.0 and 1.0.")
            elif k in CFG_INT_KEYS and not isinstance(v, int):
                raise TypeError(f"'{k}={v}' is of invalid type {type(v).__name__}. "
                                f"'{k}' must be an int (i.e. '{k}=8')")
            elif k in CFG_BOOL_KEYS and not isinstance(v, bool):
                raise TypeError(f"'{k}={v}' is of invalid type {type(v).__name__}. "
                                f"'{k}' must be a bool (i.e. '{k}=True' or '{k}=False')")

    # Return instance
    return IterableSimpleNamespace(**cfg)

def adjust_bboxes_to_image_border(boxes, image_shape, threshold=20):
    '''Adjust bounding boxes to stick to image border if they are within a certain threshold.
    Args:
    boxes: (n, 4)
    image_shape: (height, width)
    threshold: pixel threshold
    Returns:
    adjusted_boxes: adjusted bounding boxes
    '''

    # Image dimensions
    h, w = image_shape

    # Adjust boxes
    boxes[:, 0] = torch.where(boxes[:, 0] < threshold, torch.tensor(
        0, dtype=torch.float, device=boxes.device), boxes[:, 0])  # x1
    boxes[:, 1] = torch.where(boxes[:, 1] < threshold, torch.tensor(
        0, dtype=torch.float, device=boxes.device), boxes[:, 1])  # y1
    boxes[:, 2] = torch.where(boxes[:, 2] > w - threshold, torch.tensor(
        w, dtype=torch.float, device=boxes.device), boxes[:, 2])  # x2
    boxes[:, 3] = torch.where(boxes[:, 3] > h - threshold, torch.tensor(
        h, dtype=torch.float, device=boxes.device), boxes[:, 3])  # y2

    return boxes



def convert_box_xywh_to_xyxy(box):
    x1 = box[0]
    y1 = box[1]
    x2 = box[0] + box[2]
    y2 = box[1] + box[3]
    return [x1, y1, x2, y2]


def bbox_iou(box1, boxes, iou_thres=0.9, image_shape=(640, 640), raw_output=False):
    '''Compute the Intersection-Over-Union of a bounding box with respect to an array of other bounding boxes.
    Args:
    box1: (4, )
    boxes: (n, 4)
    Returns:
    high_iou_indices: Indices of boxes with IoU > thres
    '''
    boxes = adjust_bboxes_to_image_border(boxes, image_shape)
    # obtain coordinates for intersections
    x1 = torch.max(box1[0], boxes[:, 0])
    y1 = torch.max(box1[1], boxes[:, 1])
    x2 = torch.min(box1[2], boxes[:, 2])
    y2 = torch.min(box1[3], boxes[:, 3])

    # compute the area of intersection
    intersection = (x2 - x1).clamp(0) * (y2 - y1).clamp(0)

    # compute the area of both individual boxes
    box1_area = (box1[2] - box1[0]) * (box1[3] - box1[1])
    box2_area = (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])

    # compute the area of union
    union = box1_area + box2_area - intersection

    # compute the IoU
    iou = intersection / union  # Should be shape (n, )
    if raw_output:
        if iou.numel() == 0:
            return 0
        return iou

    # get indices of boxes with IoU > thres
    high_iou_indices = torch.nonzero(iou > iou_thres).flatten()

    return high_iou_indices


def image_to_np_ndarray(image):
    if type(image) is str:
        return np.array(Image.open(image))
    elif issubclass(type(image), Image.Image):
        return np.array(image)
    elif type(image) is np.ndarray:
        return image
    return None


def check_imgsz(imgsz, stride=32, min_dim=1, max_dim=2, floor=0):
    """
    Verify image size is a multiple of the given stride in each dimension. If the image size is not a multiple of the
    stride, update it to the nearest multiple of the stride that is greater than or equal to the given floor value.

    Args:
        imgsz (int | cList[int]): Image size.
        stride (int): Stride value.
        min_dim (int): Minimum number of dimensions.
        floor (int): Minimum allowed value for image size.

    Returns:
        (List[int]): Updated image size.
    """
    # Convert stride to integer if it is a tensor
    stride = int(stride.max() if isinstance(stride, torch.Tensor) else stride)

    # Convert image size to list if it is an integer
    if isinstance(imgsz, int):
        imgsz = [imgsz]
    elif isinstance(imgsz, (list, tuple)):
        imgsz = list(imgsz)
    else:
        raise TypeError(f"'imgsz={imgsz}' is of invalid type {type(imgsz).__name__}. "
                        f"Valid imgsz types are int i.e. 'imgsz=640' or list i.e. 'imgsz=[640,640]'")

    # Apply max_dim
    if len(imgsz) > max_dim:
        msg = "'train' and 'val' imgsz must be an integer, while 'predict' and 'export' imgsz may be a [h, w] list " \
              "or an integer, i.e. 'yolo export imgsz=640,480' or 'yolo export imgsz=640'"
        if max_dim != 1:
            raise ValueError(f'imgsz={imgsz} is not a valid image size. {msg}')
        logger.warning(f"WARNING ⚠️ updating to 'imgsz={max(imgsz)}'. {msg}")
        imgsz = [max(imgsz)]
    # Make image size a multiple of the stride
    sz = [max(math.ceil(x / stride) * stride, floor) for x in imgsz]

    # Print warning message if image size was updated
    if sz != imgsz:
        logger.warning(f'WARNING ⚠️ imgsz={imgsz} must be multiple of max stride {stride}, updating to {sz}')

    # Add missing dimensions if necessary
    sz = [sz[0], sz[0]] if min_dim == 2 and len(sz) == 1 else sz[0] if min_dim == 1 and len(sz) == 1 else sz

    return sz



"""AtomicOpen -- copy pasta from https://stackoverflow.com/questions/489861/locking-a-file-in-python"""
from io import FileIO
import os


try:
    # Posix based file locking (Linux, Ubuntu, MacOS, etc.)
    import fcntl

    def lock_file(f: FileIO):
        """locks a file"""
        if f.writable():
            fcntl.flock(f, fcntl.LOCK_EX)

    def unlock_file(f: FileIO):
        """unlocks a file"""
        if f.writable():
            fcntl.flock(f, fcntl.LOCK_UN)

except ModuleNotFoundError:
    # Windows file locking
    import msvcrt
    def file_size(f: FileIO):
        """file size"""
        return os.path.getsize(os.path.realpath(f.name))

    def lock_file(f: FileIO):
        """locks a file"""
        msvcrt.locking(f.fileno(), msvcrt.LK_RLCK, file_size(f))

    def unlock_file(f: FileIO):
        """unlocks a file"""
        msvcrt.locking(f.fileno(), msvcrt.LK_UNLCK, file_size(f))

class AtomicOpen:
    """
    lass for ensuring that all file operations are atomic, treat initialization like a standard call to 'open' that
    happens to be atomic. This file opener *must* be used in a "with" block.
    Open the file with arguments provided by user. Then acquire a lock on that file object.
    """
    def __init__(self, path: str, *args, **kwargs):
        # Open the file and acquire a lock on the file before operating
        self.file: FileIO = open(path, *args, **kwargs)
        # Lock the opened file
        lock_file(self.file)

    # Return the opened file object (knowing a lock has been obtained).
    def __enter__(self, *args, **kwargs):
        return self.file

    # Unlock the file and close the file object.
    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
        # Flush to make sure all buffered contents are written to file.
        self.file.flush()
        os.fsync(self.file.fileno())
        # Release the lock on the file.
        unlock_file(self.file)
        self.file.close()
        # Handle exceptions that may have come up during execution, by default any exceptions are raised to the user.
        return exc_type is None



"""Helper module to make sense of the arguments sent to vre.run()"""
from typing import Any

from .vre_video import VREVideo
from .utils import parsed_str_type
from .representations import Representation

RepresentationsSetup = dict[str, dict[str, Any]]

class VRERuntimeArgs:
    """
    VRE runtime args. Helper class to process the arguments sent to vre.run()
    Parameters:
    - video The video that this run operates on
    - representations The dictionary of representations that this run operates on
    - frames The list of frames to run VRE for. If None, will export all the frames of the video
    - exception_mode What to do when encountering an exception. It always writes the exception to 'exception.txt'
        - 'skip_representation' Will stop the run of the current representation and start the next one
        - 'stop_execution' (default) Will stop the execution of VRE
    - n_threads_data_storer The number of threads used by the DataStorer
    """
    def __init__(self, video: VREVideo, representations: list[Representation], frames: list[int] | None,
                 exception_mode: str, n_threads_data_storer: int):
        assert all(isinstance(r, Representation) for r in representations), representations
        assert exception_mode in ("stop_execution", "skip_representation"), exception_mode
        frames = sorted(list(range(len(video))) if frames is None else frames)
        assert all(isinstance(x, int) for x in frames), frames
        assert 0 <= frames[0] <= frames[-1] < len(video), f"{frames[0]=}, {frames[-1]=}, {len(video)=}"

        self.video = video
        self.frames: list[int] = frames
        self.exception_mode = exception_mode
        self.representations = representations
        self.representation_names = [r.name for r in representations]
        self.n_threads_data_storer = n_threads_data_storer

    def to_dict(self) -> dict:
        """A dict representation of this runtime args. Used in Metadata() to be stored on disk during the run."""
        return {
            "video_path": str(getattr(self.video, "path", "")),
            "video_shape": f"{self.video.shape}",
            "video_fps": f"{self.video.fps:.2f}",
            "representations": self.representation_names,
            "frames": self.frames,
            "exception_mode": self.exception_mode,
            "n_threads_data_storer": self.n_threads_data_storer,
        }

    def __repr__(self):
        return f"""[{parsed_str_type(self)}]
- Video path: '{getattr(self.video, "path", "")}'
- Representations ({len(self.representations)}): {", ".join(x for x in self.representation_names)}
- Video shape: {self.video.shape} (FPS: {self.video.fps:.2f})
- Output frames ({len(self.frames)}): [{self.frames[0]} : {self.frames[-1]}]
- Exception mode: '{self.exception_mode}'
- DataStorer Threads: {self.n_threads_data_storer} (0 = only using main thread)
"""



"""
Helper mixin class that adds the weights/device relevant methods & properties for such representation
A representation that inherits this also must have weights in the weights repository.
Note: this could arguably extend ComputeRepresentation, but we want to keep the inheritance tree as flat as possible.
"""
from abc import abstractmethod, ABC
import torch as tr

from vre.logger import vre_logger as logger

VREDevice = str | tr.device # not only torch, but this is what we support atm

class LearnedRepresentationMixin(ABC):
    """Learned Representastion Mixin for VRE implementation"""
    def __init__(self):
        self._device: VREDevice | None = None
        self.setup_called = False

    @abstractmethod
    def vre_setup(self, load_weights: bool = True):
        """
        Setup method for this representation. This is required to run this representation from within VRE. We do this
        setup separately, so we can instatiate the object without doing any VRE specific setup, like loading weights.
        """

    @abstractmethod
    def vre_free(self):
        """Needed to deallocate stuff from cuda mostly. After this, you need to run vre_setup() again."""

    @staticmethod
    @abstractmethod
    def weights_repository_links(**kwargs) -> list[str]:
        """Returns a name and a list of links to the weightss. The names are first checked in the local resources dir"""

    @property
    def device(self) -> VREDevice:
        """Returns the device of the representation"""
        if self._device is None:
            logger.warning(f"[{self}] No device set, returning 'cpu'. Call set_learned_params")
            return "cpu"
        return self._device

    @device.setter
    def device(self, dev: VREDevice):
        assert isinstance(dev, (str, tr.device)), dev
        self._device = dev

    def set_learned_params(self, **kwargs):
        """set the learned parameters for the representation"""
        attributes = ["device"]
        res = ""
        for attr in attributes:
            if attr in kwargs:
                setattr(self, attr, kwargs[attr])
                res += f"\n-{attr}: {kwargs[attr]}"

        if len(res) > 0:
            logger.debug(f"[{self}] Set node specific 'Learned' params: {res}")
